"""Bootstrap wrapper for datasets and models"""

import copy
import multiprocessing as mp
from types import MethodType
from typing import Callable, Optional, Tuple, Union

import numpy as np
from joblib import Parallel, delayed


class Boot:
    """Bootstrap wrapper for datasets and models.

    Datasets can be any sequence of numeric input, from which bootstrapped statistics
    can be calculated, with confidence intervals included.

    The models can be any model that is either callable or equipped with a `predict`
    method, such as all the models in `scikit-learn`, `pytorch` and `tensorflow`, and
    the bootstrapped model can then produce predictions with prediction intervals.

    The bootstrapped prediction intervals are computed using the an extension of method
    from [2] which also takes validation error into account. To remedy this, the .632+
    bootstrap estimate from [1] has been used. Read more in [3].

    Args:
        input (float array or model):
            Either a dataset to calculate bootstrapped statistics on, or an model for
            which bootstrapped predictions will be computed.
        random_seed (float or None):
            The random seed used for bootstrapping. If set to None then no seed will be
            set. Defaults to None.

    Examples:
        Compute the bootstrap distribution of the mean, with a 95% confidence
        interval::

            >>> from doubt.datasets import FishToxicity, PowerPlant
            >>> import numpy as np
            >>> X, y = FishToxicity().split()
            >>> boot = Boot(y, random_seed=42)
            >>> preds, intervals = boot.compute_statistic(np.mean)
            >>> round(preds, 2)
            4.06
            >>> np.around(intervals, 2)
            array([3.98, 4.17])

        Alternatively, we can output the whole bootstrap distribution::

            >>> preds, intervals = boot.compute_statistic(
            ...     np.mean,
            ...     n_boots=3,
            ...     return_all=True
            ... )
            >>> round(preds, 2)
            4.06
            >>> np.around(intervals, 2)
            array([4.06, 4.06, 4.06])

        Wrap a scikit-learn model and get prediction intervals::

            >>> from sklearn.linear_model import LinearRegression
            >>> X, y = PowerPlant().split()
            >>> linreg = Boot(LinearRegression(), random_seed=42)
            >>> linreg = linreg.fit(X, y)
            >>> preds, intervals = linreg.predict([10, 30, 1000, 50], uncertainty=0.05)
            >>> round(preds, 2)
            482.0
            >>> np.around(intervals, 2)
            array([473.5 , 490.14])

    Sources:
        [1]: Friedman, J., Hastie, T., & Tibshirani, R. (2001). The elements of
             statistical learning (Vol. 1, No. 10). New York: Springer series in
             statistics.
        [2]: Kumar, S., & Srivistava, A. N. (2012). Bootstrap prediction intervals in
             non-parametric regression with applications to anomaly detection.
        [3]: https://saattrupdan.github.io/2020-03-01-bootstrap-prediction
    """

    def __init__(self, input: object, random_seed: Optional[float] = None):
        self.random_seed = random_seed

        # Input is a model
        if callable(input) or hasattr(input, "predict"):
            self._model = input
            self.fit = MethodType(fit, self)
            self.predict = MethodType(predict, self)
            type(self).__repr__ = MethodType(_model_repr, self)  # type: ignore

        # Input is a dataset
        elif hasattr(input, "__getitem__"):
            self.data = np.asarray(input)
            self.compute_statistic = MethodType(compute_statistic, self)
            type(self).__repr__ = MethodType(_dataset_repr, self)  # type: ignore

        else:
            raise TypeError("Input not recognised.")


def _model_fit_predict(
    model, X_train: np.ndarray, y_train: np.ndarray, X_test: np.ndarray
) -> np.ndarray:
    """Fit the underlying model and perform predictions with it.

    This requires `self._model` to be set and that it is either callable or have a
    `predict` method.

    Args:
        model (object with `fit` and `predict` methods):
            The model to fit and predict with.
        X_train (float matrix):
            Feature matrix for training, of shape
            (n_train_samples, n_features).
        y_train (float array):
            Target array, of shape (n_train_samples,).
        X_test (float matrix):
            Feature matrix for predicting, of shape
            (n_test_samples, n_features).

    Returns:
        Numpy array:
            Predictions, of shape (n_test_samples,)
    """
    model = copy.deepcopy(model)
    model.fit(X_train, y_train)
    if callable(model):
        return model(X_test)
    else:
        return model.predict(X_test)


def _dataset_repr(self) -> str:
    return f"Boot(dataset_shape={self.data.shape}, " f"random_seed={self.random_seed})"


def _model_repr(self) -> str:
    model_name = self._model.__class__.__name__
    return f"Boot(model={model_name}, random_seed={self.random_seed})"


def compute_statistic(
    self,
    statistic: Callable[[np.ndarray], float],
    n_boots: Optional[int] = None,
    uncertainty: float = 0.05,
    quantiles: Optional[np.ndarray] = None,
    return_all: bool = False,
) -> Union[float, Tuple[float, np.ndarray]]:
    """Compute bootstrapped statistic.

    Args:
        statistic (callable):
            The statistic to be computed on bootstrapped samples.
        n_boots (int or None, optional):
            The number of resamples to bootstrap. If None then it is set to the square
            root of the data set. Defaults to None
        uncertainty (float, optional):
            The uncertainty used to compute the confidence interval of the bootstrapped
            statistic. Not used if `return_all` is set to True or if `quantiles` is not
            None. Defaults to 0.05.
        quantiles (Numpy array or None, optional):
            List of quantiles to output, as an alternative to the `uncertainty`
            argument, and will not be used if that argument is set. If None then
            `uncertainty` is used. Defaults to None.
        return_all (bool, optional):
            Whether all bootstrapped statistics should be returned instead of the
            confidence interval. Defaults to False.

    Returns:
        a float or a pair of a float and an array of floats:
            The statistic, and if `uncertainty` is set then also the confidence
            interval, or if `quantiles` is set then also the specified quantiles, or if
            `return_all` is set then also all of the bootstrapped statistics.
    """
    # Initialise random number generator
    rng = np.random.default_rng(self.random_seed)

    # Compute the statistic
    stat = statistic(self.data)

    # Get the number of data points
    n = int(self.data.shape[0])

    # Set default value of the number of bootstrap samples if `n_boots` is not set
    if n_boots is None:
        n_boots = int(np.sqrt(n).astype(int))

    # Compute the bootstrapped statistics
    boot_idxs = rng.choice(n, size=(n_boots, n), replace=True)
    statistics = np.apply_along_axis(statistic, 1, self.data[boot_idxs])

    if return_all:
        return stat, statistics
    else:
        # If uncertainty is set then set `quantiles` to be the two ends of the
        # confidence interval
        if uncertainty is not None:
            quantiles = [uncertainty / 2, 1.0 - (uncertainty / 2)]
        else:
            quantiles = list(quantiles)

        # Compute the quantile values
        quantile_vals = np.quantile(statistics, q=quantiles)
        return stat, quantile_vals


def predict(
    self,
    X: np.ndarray,
    n_boots: Optional[int] = None,
    uncertainty: Optional[float] = None,
    quantiles: Optional[np.ndarray] = None,
) -> Tuple[Union[float, np.ndarray], np.ndarray]:
    """Compute bootstrapped predictions.

    Args:
        X (float array):
            The array containing the data set, either of shape (f,) or (n, f), with n
            being the number of samples and f being the number of features.
        n_boots (int or None, optional):
            The number of resamples to bootstrap. If None then it is set to the square
            root of the data set. Defaults to None
        uncertainty (float or None, optional):
            The uncertainty used to compute the prediction interval of the bootstrapped
            prediction. If None then no prediction intervals are returned. Defaults to
            None.
        quantiles (sequence of floats or None, optional):
            List of quantiles to output, as an alternative to the `uncertainty`
            argument, and will not be used if that argument is set. If None then
            `uncertainty` is used. Defaults to None.

    Returns:
        float array or pair of float arrays:
            The bootstrapped predictions, and the confidence intervals if `uncertainty`
            is not None, or the specified quantiles if `quantiles` is not None.
    """
    # Initialise random number generator
    rng = np.random.default_rng(self.random_seed)

    # Ensure that input feature matrix is a Numpy array
    X = np.asarray(X)

    # If `X` is one-dimensional then expand it to two dimensions and save the
    # information, so that we can ensure the output is also one-dimensional
    onedim = len(X.shape) == 1
    if onedim:
        X = np.expand_dims(X, 0)

    # Get the full non-bootstrapped predictions of `X`
    preds = self._model(X) if callable(self._model) else self._model.predict(X)

    # If no quantiles should be outputted then simply return the predictions of the
    # underlying model
    if uncertainty is None and quantiles is None:
        return preds

    # Ensure that the underlying model has been fitted before predicting. This is only
    # a requirement if `uncertainty` is set, as we need access to `self.X_train`
    if not hasattr(self, "X_train") or self.X_train is None:
        raise RuntimeError(
            "This model has not been fitted yet! Call fit() "
            "before predicting new samples."
        )

    # Store the number of data points in the training and test datasets
    n_train = self.X_train.shape[0]

    # The authors chose the number of bootstrap samples as the square root of the
    # number of samples in the training dataset
    if n_boots is None:
        n_boots = int(np.sqrt(n_train).astype(int))

    # Sample the bootstrap indices
    train_idxs = rng.choice(n_train, size=(n_boots, n_train), replace=True)

    # Run the worker function in parallel
    with Parallel(n_jobs=mp.cpu_count() - 1) as parallel:
        bootstrap_preds_list = parallel(
            delayed(_model_fit_predict)(
                model=self._model,
                X_train=self.X_train[train_idxs[boot_idx], :],
                y_train=self.y_train[train_idxs[boot_idx]],
                X_test=X,
            )
            for boot_idx in range(n_boots)
        )

    # Convert the list of predictions to a Numpy array
    bootstrap_preds = np.array(bootstrap_preds_list)

    # Centre the bootstrapped predictions across the bootstrap dimension
    bootstrap_preds -= np.mean(bootstrap_preds, axis=0)

    # Add up the bootstrap predictions and the hybrid train/val residuals
    C = np.array([m + o for m in bootstrap_preds for o in self.residuals])

    # Calculate the desired quantiles
    if quantiles is None and uncertainty is not None:
        quantiles = [uncertainty / 2, 1 - uncertainty / 2]
    quantile_vals = np.transpose(np.quantile(C, q=quantiles or [], axis=0))

    # Return the predictions and the desired quantiles
    if onedim:
        return preds[0], (preds + quantile_vals)[0]
    else:
        return preds, np.expand_dims(preds, axis=1) + quantile_vals


def fit(self, X: np.ndarray, y: np.ndarray, n_boots: Optional[int] = None):
    """Fits the model to the data.

    Args:
        X (float array):
            The array containing the data set, either of shape (f,) or (n, f), with n
            being the number of samples and f being the number of features.
        y (float array):
            The array containing the target values, of shape (n,)
        n_boots (int or None):
            The number of resamples to bootstrap. If None then it is set to the square
            root of the data set. Defaults to None
    """
    # Initialise random number generator
    rng = np.random.default_rng(self.random_seed)

    # Set the number of data points in the dataset
    n = X.shape[0]

    # Set default value of `n_boots` if it is not set
    if n_boots is None:
        n_boots = int(np.sqrt(n).astype(int))

    # Ensure that `X` and `y` are Numpy arrays
    X = np.asarray(X)
    y = np.asarray(y)

    # Store `X` and `y` for predictions
    self.X_train = X
    self.y_train = y

    # Fit the underlying model and get predictions on the training dataset
    self._model.fit(X, y)
    preds = self._model(X) if callable(self._model) else self._model.predict(X)

    # Calculate the training residuals and aggregate them into quantiles, to enable
    # comparison with the validation residuals
    train_residuals = np.quantile(y - preds, q=np.arange(0, 1, 0.01))

    # Sample the bootstrap indices
    train_idxs = rng.choice(n, size=(n_boots, n), replace=True)

    # Run the worker function in parallel
    with Parallel(n_jobs=mp.cpu_count() - 1) as parallel:
        bootstrap_preds = parallel(
            delayed(_model_fit_predict)(
                model=self._model,
                X_train=self.X_train[train_idxs[boot_idx], :],
                y_train=self.y_train[train_idxs[boot_idx]],
                X_test=X[[idx for idx in range(n) if idx not in train_idxs[boot_idx]]],
            )
            for boot_idx in range(n_boots)
        )

    # Compute the validation residuals
    val_residuals_list = [
        y[[idx for idx in range(n) if idx not in train_idxs[boot_idx]]]
        - bootstrap_preds[boot_idx]
        for boot_idx in range(n_boots)
    ]

    # Aggregate the validation residuals into quantiles, to enable comparison with the
    # training residuals
    val_residuals = np.concatenate(val_residuals_list)
    val_residuals = np.quantile(val_residuals, q=np.arange(0, 1, 0.01))

    # Compute the no-information value
    permuted = rng.permutation(y) - rng.permutation(preds)
    no_info_error = np.mean(np.abs(permuted))
    no_info_val = np.abs(no_info_error - train_residuals)

    # Compute the .632+ bootstrap estimate for the sample noise and bias
    generalisation = np.abs(val_residuals.mean() - train_residuals.mean())
    relative_overfitting_rate = np.mean(generalisation / no_info_val)
    weight = 0.632 / (1 - 0.368 * relative_overfitting_rate)
    self.residuals = (1 - weight) * train_residuals + weight * val_residuals

    return self



"""Power plant data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io
import zipfile

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class PowerPlant(BaseDataset):
    __doc__ = f"""
    The dataset contains 9568 data points collected from a Combined Cycle Power Plant
    over 6 years (2006-2011), when the power plant was set to work with full load.
    Features consist of hourly average ambient variables Temperature (T), Ambient
    Pressure (AP), Relative Humidity (RH) and Exhaust Vacuum (V) to predict the net
    hourly electrical energy output (EP) of the plant.

    A combined cycle power plant (CCPP) is composed of gas turbines (GT), steam
    turbines (ST) and heat recovery steam generators. In a CCPP, the electricity is
    generated by gas and steam turbines, which are combined in one cycle, and is
    transferred from one turbine to another. While the Vacuum is colected from and has
    effect on the Steam Turbine, he other three of the ambient variables effect the GT
    performance.

    For comparability with our baseline studies, and to allow 5x2 fold statistical
    tests be carried out, we provide the data shuffled five times. For each shuffling
    2-fold CV is carried out and the resulting 10 measurements are used for statistical
    testing.

    {BASE_DATASET_DESCRIPTION}

    Features:
        AT (float):
            Hourly average temperature in Celsius, ranges from 1.81 to 37.11
        V (float):
            Hourly average exhaust vacuum in cm Hg, ranges from 25.36 to 81.56
        AP (float):
            Hourly average ambient pressure in millibar, ranges from 992.89
            to 1033.30
        RH (float):
            Hourly average relative humidity in percent, ranges from 25.56 to 100.16

    Targets:
        PE (float):
            Net hourly electrical energy output in MW, ranges from 420.26 to 495.76

    Source:
        https://archive.ics.uci.edu/ml/datasets/Combined+Cycle+Power+Plant

    Examples:
        Load in the data set::

            >>> dataset = PowerPlant()
            >>> dataset.shape
            (9568, 5)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((9568, 4), (9568,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((7633, 4), (7633,), (1935, 4), (1935,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = "https://archive.ics.uci.edu/ml/machine-learning-databases/" "00294/CCPP.zip"

    _features = range(4)
    _targets = [4]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """

        # Convert the bytes into a file-like object
        buffer = io.BytesIO(data)

        # Unzip the file and pull out the xlsx file
        with zipfile.ZipFile(buffer, "r") as zip_file:
            xlsx = zip_file.read("CCPP/Folds5x2_pp.xlsx")

        # Convert the xlsx bytes into a file-like object
        xlsx_file = io.BytesIO(xlsx)

        # Read the file-like object into a dataframe
        df = pd.read_excel(xlsx_file)
        return df



"""CPU data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class CPU(BaseDataset):
    __doc__ = f"""
    Relative CPU Performance Data, described in terms of its cycle time, memory size,
    etc.

    {BASE_DATASET_DESCRIPTION}

    Features:
        vendor_name (string):
            Name of the vendor, 30 unique values
        model_name (string):
            Name of the model
        myct (int):
            Machine cycle time in nanoseconds
        mmin (int):
            Minimum main memory in kilobytes
        mmax (int):
            Maximum main memory in kilobytes
        cach (int):
            Cache memory in kilobytes
        chmin (int):
            Minimum channels in units
        chmax (int):
            Maximum channels in units

    Targets:
        prp (int):
            Published relative performance

    Source:
        https://archive.ics.uci.edu/ml/datasets/Computer+Hardware

    Examples:
        Load in the data set::

            >>> dataset = CPU()
            >>> dataset.shape
            (209, 9)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((209, 8), (209,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((162, 8), (162,), (47, 8), (47,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "cpu-performance/machine.data"
    )

    _features = range(8)
    _targets = [8]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """

        # Convert the bytes into a file-like object
        csv_file = io.BytesIO(data)

        # Name the columns
        cols = [
            "vendor_name",
            "model_name",
            "myct",
            "mmin",
            "mmax",
            "cach",
            "chmin",
            "chmax",
            "prp",
        ]

        # Load the file-like object into a data frame
        df = pd.read_csv(csv_file, header=None, usecols=range(9), names=cols)
        return df



"""New Taipei Housing data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class NewTaipeiHousing(BaseDataset):
    __doc__ = f"""
    The "real estate valuation" is a regression problem. The market historical data set
    of real estate valuation are collected from Sindian Dist., New Taipei City, Taiwan.

    {BASE_DATASET_DESCRIPTION}

    Features:
        transaction_date (float):
            The transaction date encoded as a floating point value. For instance,
            2013.250 is March 2013 and 2013.500 is June March
        house_age (float):
            The age of the house
        mrt_distance (float):
            Distance to the nearest MRT station
        n_stores (int):
            Number of convenience stores
        lat (float):
            Latitude
        lng (float):
            Longitude

    Targets:
        house_price (float):
            House price of unit area

    Source:
        https://archive.ics.uci.edu/ml/datasets/Real+estate+valuation+data+set

    Examples:
        Load in the data set::

            >>> dataset = NewTaipeiHousing()
            >>> dataset.shape
            (414, 7)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((414, 6), (414,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((323, 6), (323,), (91, 6), (91,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00477/Real%20estate%20valuation%20data%20set.xlsx"
    )

    _features = range(6)
    _targets = [6]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        xlsx_file = io.BytesIO(data)

        # Load in the dataframe
        cols = [
            "idx",
            "transaction_date",
            "house_age",
            "mrt_distance",
            "n_stores",
            "lat",
            "lng",
            "house_price",
        ]
        df = pd.read_excel(xlsx_file, header=0, names=cols)

        # Remove the index
        df = df.iloc[:, 1:]

        return df



"""Utility functions used in tree models"""

from typing import Optional

import numpy as np


def weighted_percentile(
    arr: np.ndarray,
    quantile: float,
    weights: Optional[np.ndarray] = None,
    sorter: Optional[np.ndarray] = None,
):
    """Returns the weighted percentile of an array.

    See [1] for an explanation of this concept.

    Args:
        arr (array-like):
            Samples at which the quantile should be computed, of shape [n_samples,].
        quantile (float):
            Quantile, between 0.0 and 1.0.
        weights (array-like, optional):
            The weights, of shape = (n_samples,). Here weights[i] is the weight given
            to point a[i] while computing the quantile. If weights[i] is zero, a[i] is
            simply ignored during the percentile computation. If None then uniform
            weights will be used. Defaults to None.
        sorter (array-like, optional):
            Array of shape [n_samples,], indicating the indices sorting `arr`. Thus, if
            provided, we assume that arr[sorter] is sorted. If None then `arr` will be
            sorted. Defaults to None.

    Returns:
        percentile: float
            Weighted percentile of `arr` at `quantile`.

    Raises:
        ValueError:
            If `quantile` is not between 0.0 and 1.0, or if `arr` and `weights` are of
            different lengths.

    Sources:
        [1]: https://en.wikipedia.org/wiki/Percentile#The_weighted_percentile_method
    """
    # Ensure that quantile is set properly
    if quantile > 1 or quantile < 0:
        raise ValueError("The quantile should be between 0 and 1.")

    # Set weights to be uniform if not specified
    weights_arr = np.ones_like(arr) if weights is None else weights

    # Ensure that `arr` and `weights_arr` are numpy arrays
    arr = np.asarray(arr, dtype=np.float32)
    weights = np.asarray(weights_arr, dtype=np.float32)

    # Ensure that `arr` and `weights` are of the same length
    if len(arr) != len(weights_arr):
        raise ValueError("`arr` and `weights` should have the same length.")

    # If `sorter` is given , then sort `arr` and `weights_arr` using it
    if sorter is not None:
        arr = arr[sorter]
        weights_arr = weights_arr[sorter]

    # Remove all the array (and weight) elements with zero weight
    non_zeros = weights_arr != 0
    arr = arr[non_zeros]
    weights_arr = weights_arr[non_zeros]

    # Sort the array if `sorter` is not given
    if sorter is None:
        sorted_indices = np.argsort(arr)
        sorted_arr = arr[sorted_indices]
        sorted_weights = weights_arr[sorted_indices]
    else:
        sorted_arr = arr
        sorted_weights = weights_arr

    # Calculate the partial sum of weights and get the total weight
    sorted_cum_weights = np.cumsum(sorted_weights)
    total = sorted_cum_weights[-1]

    # Calculate the percentile values
    partial_sum = 1.0 / total
    partial_sum *= sorted_cum_weights - sorted_weights / 2.0

    # Find the spot in `partial_sum` where `quantile` belongs, preserving order
    start = np.searchsorted(partial_sum, quantile) - 1

    # If the spot is the first or last, return the first or last percentiles
    if start == len(sorted_cum_weights) - 1:
        return sorted_arr[-1]
    if start == -1:
        return sorted_arr[0]

    # Find the proportion of which the distance from `partial_sum[start]` to `quantile`
    # is compared to the distance from `partial_sum[start]` to `partial_sum[start + 1]`
    fraction = quantile - partial_sum[start]
    fraction /= partial_sum[start + 1] - partial_sum[start]

    # Add the corresponding proportion from `sorted_arr[start]` to
    # `sorted_arr[start + 1]`, to `sorted_arr[start]`.
    return sorted_arr[start] + fraction * (sorted_arr[start + 1] - sorted_arr[start])



"""Facebook comments data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io
import zipfile

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class FacebookComments(BaseDataset):
    __doc__ = f"""
    Instances in this dataset contain features extracted from Facebook posts. The task
    associated with the data is to predict how many comments the post will receive.

    {BASE_DATASET_DESCRIPTION}

    Features:
        page_popularity (int):
            Defines the popularity of support for the source of the document
        page_checkins (int):
            Describes how many individuals so far visited this place. This feature is
            only associated with places; e.g., some institution, place, theater, etc.
        page_talking_about (int):
            Defines the daily interest of individuals towards source of the
            document/post. The people who actually come back to the page, after liking
            the page. This include activities such as comments, likes to a post, shares
            etc., by visitors to the page
        page_category (int):
            Defines the category of the source of the document; e.g., place,
            institution, branch etc.
        agg[n] for n=0..24 (float):
            These features are aggreagted by page, by calculating min, max, average,
            median and standard deviation of essential features
        cc1 (int):
            The total number of comments before selected base date/time
        cc2 (int):
            The number of comments in the last 24 hours, relative to base date/time
        cc3 (int):
            The number of comments in the last 48 to last 24 hours relative to base
            date/time
        cc4 (int):
            The number of comments in the first 24 hours after the publication of post
            but before base date/time
        cc5 (int):
            The difference between cc2 and cc3
        base_time (int):
            Selected time in order to simulate the scenario, ranges from 0 to 71
        post_length (int):
            Character count in the post
        post_share_count (int):
            This feature counts the number of shares of the post, how many people had
            shared this post onto their timeline
        post_promotion_status (int):
            Binary feature. To reach more people with posts in News Feed, individuals
            can promote their post and this feature indicates whether the post is
            promoted or not
        h_local (int):
            This describes the hours for which we have received the target
            variable/comments. Ranges from 0 to 23
        day_published[n] for n=0..6 (int):
            Binary feature. This represents the day (Sunday-Saturday) on which the post
            was published
        day[n] for n=0..6 (int):
            Binary feature. This represents the day (Sunday-Saturday) on selected base
            date/time

    Targets:
        ncomments (int):
            The number of comments in the next `h_local` hours

    Source:
        https://archive.ics.uci.edu/ml/datasets/Facebook+Comment+Volume+Dataset

    Examples:
        Load in the data set::

            >>> dataset = FacebookComments()
            >>> dataset.shape
            (199030, 54)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((199030, 54), (199030,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((159211, 54), (159211,), (39819, 54), (39819,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/" "00363/Dataset.zip"
    )

    _features = range(54)
    _targets = [53]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """

        # Convert the bytes into a file-like object
        buffer = io.BytesIO(data)

        # Unzip the file and pull out the csv file
        with zipfile.ZipFile(buffer, "r") as zip_file:
            csv = zip_file.read("Dataset/Training/Features_Variant_5.csv")

        # Convert the string into a file-like object
        csv_file = io.BytesIO(csv)

        # Name the columns
        cols = (
            ["page_popularity", "page_checkins", "page_talking_about", "page_category"]
            + [f"agg{n}" for n in range(25)]
            + [
                "cc1",
                "cc2",
                "cc3",
                "cc4",
                "cc5",
                "base_time",
                "post_length",
                "post_share_count",
                "post_promotion_status",
                "h_local",
            ]
            + [f"day_published{n}" for n in range(7)]
            + [f"day{n}" for n in range(7)]
            + ["ncomments"]
        )

        # Read the file-like object into a dataframe
        df = pd.read_csv(csv_file, header=None, names=cols)
        return df



"""Quantile regression forests"""

from typing import Optional, Tuple, Union

import numpy as np
from joblib import Parallel, delayed
from tqdm.auto import tqdm

from .tree import QuantileRegressionTree


class QuantileRegressionForest:
    """A random forest for regression which can output quantiles as well.

    Args:
        n_estimators (int, optional):
            The number of trees in the forest. Defaults to 100.
        criterion (string, optional):
            The function to measure the quality of a split. Supported criteria are
            'squared_error' for the mean squared error, which is equal to variance
            reduction as feature selection criterion, and 'absolute_error' for the mean
            absolute error. Defaults to 'squared_error'.
        splitter (string, optional):
            The strategy used to choose the split at each node. Supported strategies
            are 'best' to choose the best split and 'random' to choose the best random
            split. Defaults to 'best'.
        max_features (int, float, string or None, optional):
            The number of features to consider when looking for the best split:

            - If int, then consider `max_features` features at each split.
            - If float, then `max_features` is a percentage and `int(max_features *
              n_features)` features are considered at each split.
            - If 'auto', then `max_features=n_features`.
            - If 'sqrt', then `max_features=sqrt(n_features)`.
            - If 'log2', then `max_features=log2(n_features)`.
            - If None, then `max_features=n_features`.

            Note: the search for a split does not stop until at least one valid
            partition of the node samples is found, even if it requires to effectively
            inspect more than ``max_features`` features. Defaults to None.
        max_depth (int or None, optional):
            The maximum depth of the tree. If None, then nodes are expanded until all
            leaves are pure or until all leaves contain less than min_samples_split
            samples. Defaults to None.
        min_samples_split (int or float, optional):
            The minimum number of samples required to split an internal node:

            - If int, then consider `min_samples_split` as the minimum number.
            - If float, then `min_samples_split` is a percentage and
              `ceil(min_samples_split * n_samples)` are the minimum number of samples
              for each split. Defaults to 2.

        min_samples_leaf (int or float, optional):
            The minimum number of samples required to be at a leaf node:

            - If int, then consider `min_samples_leaf` as the minimum number.
            - If float, then `min_samples_leaf` is a percentage and
              `ceil(min_samples_leaf * n_samples)` are the minimum number of samples
              for each node. Defaults to 5.

        min_weight_fraction_leaf (float, optional):
            The minimum weighted fraction of the sum total of weights (of all the input
            samples) required to be at a leaf node. Samples have equal weight when
            sample_weight is not provided. Defaults to 0.0.
        max_leaf_nodes (int or None, optional):
            Grow a tree with ``max_leaf_nodes`` in best-first fashion. Best nodes are
            defined as relative reduction in impurity. If None then unlimited number of
            leaf nodes. Defaults to None.
        n_jobs (int, optional):
            The number of CPU cores used in fitting and predicting. If -1 then all
            available CPU cores will be used. Defaults to -1.
        random_seed (int, RandomState instance or None, optional):
            If int, random_state is the seed used by the random number generator; If
            RandomState instance, random_state is the random number generator; If None,
            the random number generator is the RandomState instance used by
            `np.random`. Defaults to None.
        verbose (bool, optional):
            Whether extra output should be printed during training and inference.
            Defaults to False.

    Examples:
        Fitting and predicting follows scikit-learn syntax::

            >>> from doubt.datasets import Concrete
            >>> import numpy as np
            >>> X, y = Concrete().split()
            >>> forest = QuantileRegressionForest(random_seed=42, max_leaf_nodes=8)
            >>> forest.fit(X, y).predict(X).shape
            (1030,)
            >>> preds = forest.predict(np.ones(8))
            >>> 16 < preds < 17
            True

        Instead of only returning the prediction, we can also return a prediction
        interval::

            >>> preds, interval = forest.predict(np.ones(8), uncertainty=0.25)
            >>> interval[0] < preds < interval[1]
            True

        We can also get the feature importances for the model::

            >>> np.round(forest.feature_importances_, 1)
            array([0.4, 0. , 0. , 0.1, 0.1, 0. , 0. , 0.4])
    """

    def __init__(
        self,
        n_estimators: int = 100,
        criterion: str = "squared_error",
        splitter: str = "best",
        max_features: Optional[Union[int, float, str]] = None,
        max_depth: Optional[int] = None,
        min_samples_split: Union[int, float] = 2,
        min_samples_leaf: Union[int, float] = 5,
        min_weight_fraction_leaf: float = 0.0,
        max_leaf_nodes: Optional[int] = None,
        n_jobs: int = -1,
        random_seed: Optional[int] = None,
        verbose: bool = False,
    ):

        self.n_estimators = n_estimators
        self.min_samples_leaf = min_samples_leaf
        self.criterion = criterion
        self.splitter = splitter
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.n_jobs = n_jobs
        self.random_seed = random_seed
        self.verbose = verbose

        self.estimators_ = n_estimators * [
            QuantileRegressionTree(
                criterion=criterion,
                splitter=splitter,
                max_depth=max_depth,
                min_samples_split=min_samples_split,
                min_samples_leaf=min_samples_leaf,
                min_weight_fraction_leaf=min_weight_fraction_leaf,
                max_features=max_features,
                max_leaf_nodes=max_leaf_nodes,
                random_seed=random_seed,
            )
        ]

    @property
    def feature_importances_(self) -> np.ndarray:
        """The impurity-based feature importances.

        The higher, the more important the feature. The importance of a feature is
        computed as the (normalized) total reduction of the criterion brought by that
        feature.  It is also known as the Gini importance. Warning: impurity-based
        feature importances can be misleading for high cardinality features (many
        unique values).

        Returns:
            ndarray of shape (n_features,):
                The values of this array sum to 1, unless all trees are single node
                trees consisting of only the root node, in which case it will be an
                array of zeros.
        """
        all_importances = Parallel(n_jobs=self.n_jobs, prefer="threads")(
            delayed(getattr)(tree, "feature_importances_")
            for tree in self.estimators_
            if tree.tree_.node_count > 1
        )
        all_importances = np.mean(all_importances, axis=0, dtype=np.float64)
        return all_importances / np.sum(all_importances)

    def __repr__(self) -> str:
        txt = "QuantileRegressionForest("
        attributes = [
            "n_estimators",
            "criterion",
            "splitter",
            "max_features",
            "max_depth",
            "min_samples_split",
            "min_samples_leaf",
            "min_weight_fraction_leaf",
            "max_leaf_nodes",
            "n_jobs",
            "random_seed",
        ]
        for idx, attr in enumerate(attributes):
            if idx > 0:
                txt += "                         "
            txt += f"{attr}={getattr(self, attr)}"
            if idx < len(attributes) - 1:
                txt += ",\n"
        return txt + ")"

    def fit(self, X, y, **kwargs):
        """Fit decision trees in parallel.

        Args:
            X (array-like or sparse matrix):
                The input samples, of shape [n_samples, n_features]. Internally, it
                will be converted to `dtype=np.float32` and if a sparse matrix is
                provided to a sparse `csr_matrix`.
            y (array-like):
                The target values (class labels) as integers or strings, of shape
                [n_samples] or [n_samples, n_outputs].
            verbose (bool or None, optional):
                Whether extra output should be printed during training. If None then
                the initialised value of the `verbose` parameter will be used. Defaults
                to None.
        """
        # Set the verbose argument if it has not been set
        if kwargs.get("verbose") is None:
            verbose = self.verbose
        else:
            verbose = kwargs.get("verbose")

        # Initialise random number generator
        rng = np.random.default_rng(self.random_seed)

        # Store the number of training samples
        n = X.shape[0]

        # Get bootstrap resamples of the data set
        bidxs = rng.choice(n, size=(self.n_estimators, n), replace=True)

        # Set up progress bar if requested
        if verbose:
            itr = tqdm(self.estimators_, desc="Fitting trees")
        else:
            itr = self.estimators_

        # Fit trees in parallel on the bootstrapped resamples
        with Parallel(n_jobs=self.n_jobs) as parallel:
            self.estimators_ = parallel(
                delayed(estimator.fit)(X[bidxs[b, :], :], y[bidxs[b, :]])
                for b, estimator in enumerate(itr)
            )

        if verbose:
            itr.close()

        return self

    def predict(
        self,
        X: np.ndarray,
        uncertainty: Optional[float] = None,
        quantiles: Optional[np.ndarray] = None,
        verbose: Optional[bool] = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Predict regression value for X.

        Args:
            X (array-like or sparse matrix):
                The input samples, of shape [n_samples, n_features]. Internally, it
                will be converted to `dtype=np.float32` and if a sparse matrix is
                provided to a sparse `csr_matrix`.
            uncertainty (float or None, optional):
                Value ranging from 0 to 1. If None then no prediction intervals will be
                returned. Defaults to None.
            quantiles (sequence of floats or None, optional):
                List of quantiles to output, as an alternative to the `uncertainty`
                argument, and will not be used if that argument is set. If None then
                `uncertainty` is used. Defaults to None.
            verbose (bool or None, optional):
                Whether extra output should be printed during inference. If None then
                the initialised value of the `verbose` parameter will be used. Defaults
                to None.

        Returns:
            Array or pair of arrays:
                Either array with predictions, of shape [n_samples,], or a pair of
                arrays with the first one being the predictions and the second one
                being the desired quantiles/intervals, of shape [2, n_samples] if
                `uncertainty` is not None, and [n_quantiles, n_samples] if `quantiles`
                is not None.
        """
        # Set the verbose argument if it has not been set
        if verbose is None:
            verbose = self.verbose

        # Ensure that X is two-dimensional
        onedim = len(X.shape) == 1
        if onedim:
            X = np.expand_dims(X, 0)

        # Set up progress bar if requested
        if verbose:
            itr = tqdm(self.estimators_, desc="Getting tree predictions")
        else:
            itr = self.estimators_

        with Parallel(n_jobs=self.n_jobs) as parallel:
            preds = parallel(
                delayed(estimator.predict)(
                    X, uncertainty=uncertainty, quantiles=quantiles
                )
                for estimator in itr
            )

            if uncertainty is not None or quantiles is not None:
                quantile_vals = np.stack([interval for _, interval in preds], axis=0)
                quantile_vals = quantile_vals.mean(0)
                preds = np.stack([pred for pred, _ in preds])
                preds = preds.mean(0)
                if onedim:
                    preds = preds[0]
                    quantile_vals = quantile_vals[0]
                return preds, quantile_vals

            else:
                preds = np.mean(preds, axis=0)
                if onedim:
                    preds = preds[0]
                return preds



"""Implementation of the quantile loss function"""

import numpy as np


def quantile_loss(
    predictions: np.ndarray, targets: np.ndarray, quantile: float
) -> float:
    """Quantile loss function.

    Args:
        predictions (sequence of floats):
            Model predictions, of shape [n_samples,].
        targets (sequence of floats):
            Target values, of shape [n_samples,].
        quantile (float):
            The quantile we are seeking. Must be between 0 and 1.

    Returns:
        float: The quantile loss.
    """
    # Convert inputs to NumPy arrays
    target_arr = np.asarray(targets)
    prediction_arr = np.asarray(predictions)

    # Compute the residuals
    res = target_arr - prediction_arr

    # Compute the mean quantile loss
    loss = np.mean(
        np.maximum(res, np.zeros_like(res)) * quantile
        + np.maximum(-res, np.zeros_like(res)) * (1 - quantile)
    )

    # Ensure that loss is of type float and return it
    return float(loss)


def smooth_quantile_loss(
    predictions: np.ndarray,
    targets: np.ndarray,
    quantile: float,
    alpha: float = 0.4,
) -> float:
    """The smooth quantile loss function from [1].

    Args:
        predictions (sequence of floats):
            Model predictions, of shape [n_samples,].
        targets (sequence of floats):
            Target values, of shape [n_samples,].
        quantile (float):
            The quantile we are seeking. Must be between 0 and 1.
        alpha (float, optional):
            Smoothing parameter. Defaults to 0.4.

    Returns:
        float: The smooth quantile loss.

    Sources:
        [1]: Songfeng Zheng (2011). Gradient Descent Algorithms for Quantile Regression
             With Smooth Approximation. International Journal of Machine Learning and
             Cybernetics.
    """
    # Convert inputs to NumPy arrays
    target_arr = np.asarray(targets)
    prediction_arr = np.asarray(predictions)

    # Compute the residuals
    residuals = target_arr - prediction_arr

    # Compute the smoothened mean quantile loss
    loss = quantile * residuals + alpha * np.log(1 + np.exp(-residuals / alpha))

    # Ensure that loss is of type float and return it
    return float(loss.mean())



"""Base class for data sets"""

import re
import warnings
from abc import ABC, abstractmethod
from typing import Iterable, Optional, Tuple, Union

import numpy as np
import pandas as pd
import requests

BASE_DATASET_DESCRIPTION = """
    Parameters:
        cache (str or None, optional):
            The name of the cache. It will be saved to `cache` in the current working
            directory. If None then no cache will be saved. Defaults to
            '.dataset_cache'.

    Attributes:
        cache (str or None):
            The name of the cache.
        shape (tuple of integers):
            Dimensions of the data set
        columns (list of strings):
            List of column names in the data set
"""


class BaseDataset(ABC):

    _url: str
    _features: Iterable
    _targets: Iterable

    def __init__(self, cache: Optional[str] = ".dataset_cache"):
        self.cache = cache
        self._data = self.get_data()
        self.shape = self._data.shape
        self.columns = self._data.columns

    @abstractmethod
    def _prep_data(self, data: bytes) -> pd.DataFrame:
        return

    def get_data(self) -> pd.DataFrame:
        """Download and prepare the dataset.

        Returns:
            Pandas DataFrame: The dataset.
        """

        # Get name of dataset, being the class name converted to snake case
        name = re.sub(r"([A-Z])", r"_\1", type(self).__name__)
        name = name.lower().strip("_")

        try:
            if self.cache is not None:
                data = pd.read_hdf(self.cache, name)
        except (FileNotFoundError, KeyError):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                response = requests.get(self._url, verify=False)
            data = self._prep_data(response.content)
            if self.cache is not None:
                data.to_hdf(self.cache, name)
        return data

    def to_pandas(self) -> pd.DataFrame:
        return self._data

    def __len__(self) -> int:
        return len(self._data)

    def head(self, n: int = 5) -> pd.DataFrame:
        return self._data.head(n)

    def close(self):
        del self._data
        del self

    def __exit__(self, exc_type: str, exc_value: str, exc_traceback: str):
        self.close()

    def __str__(self) -> str:
        return str(self._data)

    def __repr__(self) -> str:
        return repr(self._data)

    def _repr_html_(self):
        return self._data._repr_html_()

    def split(
        self, test_size: Optional[float] = None, random_seed: Optional[float] = None
    ) -> Union[
        Tuple[np.ndarray, np.ndarray],
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
    ]:
        """Split dataset into features and targets and optionally train/test.

        Args:
            test_size (float or None):
                The fraction of the dataset that will constitute the test set. If None
                then no train/test split will happen. Defaults to None.
            random_seed (float or None):
                The random seed used for the train/test split. If None then a random
                number will be chosen. Defaults to None.

        Returns:
            If `test_size` is not `None` then a tuple of numpy arrays
            (X_train, y_train, X_test, y_test) is returned, and otherwise
            the tuple (X, y) of numpy arrays is returned.
        """
        # Initialise random number generator
        rng = np.random.default_rng(random_seed)

        nrows = len(self._data)
        features = self._features
        targets = self._targets

        if test_size is not None:
            test_idxs = rng.random(size=(nrows,)) < test_size
            train_idxs = ~test_idxs

            X_train = self._data.iloc[train_idxs, features].values
            y_train = self._data.iloc[train_idxs, targets].values.squeeze()
            X_test = self._data.iloc[test_idxs, features].values
            y_test = self._data.iloc[test_idxs, targets].values.squeeze()

            return X_train, X_test, y_train, y_test

        else:
            X = self._data.iloc[:, features].values
            y = self._data.iloc[:, targets].values.squeeze()
            return X, y



"""Stocks data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class Stocks(BaseDataset):
    __doc__ = f"""
    There are three disadvantages of weighted scoring stock selection models. First,
    they cannot identify the relations between weights of stock-picking concepts and
    performances of portfolios. Second, they cannot systematically discover the optimal
    combination for weights of concepts to optimize the performances. Third, they are
    unable to meet various investors' preferences.

    This study aims to more efficiently construct weighted scoring stock selection
    models to overcome these disadvantages. Since the weights of stock-picking concepts
    in a weighted scoring stock selection model can be regarded as components in a
    mixture, we used the simplex centroid mixture design to obtain the experimental
    sets of weights. These sets of weights are simulated with US stock market
    historical data to obtain their performances. Performance prediction models were
    built with the simulated performance data set and artificial neural networks.

    Furthermore, the optimization models to reflect investors' preferences were built
    up, and the performance prediction models were employed as the kernel of the
    optimization models so that the optimal solutions can now be solved with
    optimization techniques. The empirical values of the performances of the optimal
    weighting combinations generated by the optimization models showed that they can
    meet various investors' preferences and outperform those of S&P's 500 not only
    during the training period but also during the testing period.

    {BASE_DATASET_DESCRIPTION}

    Features:
        bp (float):
            Large B/P
        roe (float):
            Large ROE
        sp (float):
            Large S/P
        return_rate (float):
            Large return rate in the last quarter
        market_value (float):
            Large market value
        small_risk (float):
            Small systematic risk
        orig_annual_return (float):
            Annual return
        orig_excess_return (float):
            Excess return
        orig_risk (float):
            Systematic risk
        orig_total_risk (float):
            Total risk
        orig_abs_win_rate (float):
            Absolute win rate
        orig_rel_win_rate (float):
            Relative win rate

    Targets:
        annual_return (float):
            Annual return
        excess_return (float):
            Excess return
        risk (float):
            Systematic risk
        total_risk (float):
            Total risk
        abs_win_rate (float):
            Absolute win rate
        rel_win_rate (float):
            Relative win rate

    Source:
        https://archive.ics.uci.edu/ml/datasets/Stock+portfolio+performance

    Examples:
        Load in the data set::

            >>> dataset = Stocks()
            >>> dataset.shape
            (252, 19)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((252, 12), (252, 6))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((197, 12), (197, 6), (55, 12), (55, 6))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00390/stock%20portfolio%20performance%20data%20set.xlsx"
    )

    _features = range(12)
    _targets = range(12, 18)

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        xlsx_file = io.BytesIO(data)

        # Load in the dataframes
        cols = [
            "id",
            "bp",
            "roe",
            "sp",
            "return_rate",
            "market_value",
            "small_risk",
            "orig_annual_return",
            "orig_excess_return",
            "orig_risk",
            "orig_total_risk",
            "orig_abs_win_rate",
            "orig_rel_win_rate",
            "annual_return",
            "excess_return",
            "risk",
            "total_risk",
            "abs_win_rate",
            "rel_win_rate",
        ]
        sheets = ["1st period", "2nd period", "3rd period", "4th period"]
        dfs = pd.read_excel(
            xlsx_file, sheet_name=sheets, names=cols, skiprows=[0, 1], header=None
        )

        # Concatenate the dataframes
        df = pd.concat([dfs[sheet] for sheet in sheets], ignore_index=True)

        return df



"""Nanotube data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class Nanotube(BaseDataset):
    __doc__ = f"""
    CASTEP can simulate a wide range of properties of materials proprieties using
    density functional theory (DFT). DFT is the most successful method calculates
    atomic coordinates faster than other mathematical approaches, and it also reaches
    more accurate results. The dataset is generated with CASTEP using CNT geometry
    optimization. Many CNTs are simulated in CASTEP, then geometry optimizations are
    calculated. Initial coordinates of all carbon atoms are generated randomly.
    Different chiral vectors are used for each CNT simulation.

    The atom type is selected as carbon, bond length is used as 1.42 AÂ° (default
    value). CNT calculation parameters are used as default parameters. To finalize the
    computation, CASTEP uses a parameter named as elec_energy_tol (electrical energy
    tolerance) (default 1x10-5 eV) which represents that the change in the total energy
    from one iteration to the next remains below some tolerance value per atom for a
    few self-consistent field steps. Initial atomic coordinates (u, v, w), chiral
    vector (n, m) and calculated atomic coordinates (u, v, w) are obtained from the
    output files.

    {BASE_DATASET_DESCRIPTION}

    Features:
        Chiral indice n (int):
            n parameter of the selected chiral vector
        Chiral indice m (int):
            m parameter of the selected chiral vector
        Initial atomic coordinate u (float):
            Randomly generated u parameter of the initial atomic coordinates
            of all carbon atoms.
        Initial atomic coordinate v (float):
            Randomly generated v parameter of the initial atomic coordinates
            of all carbon atoms.
        Initial atomic coordinate w (float):
            Randomly generated w parameter of the initial atomic coordinates
            of all carbon atoms.

    Targets:
        Calculated atomic coordinates u (float):
           Calculated u parameter of the atomic coordinates of all
           carbon atoms
        Calculated atomic coordinates v (float):
           Calculated v parameter of the atomic coordinates of all
           carbon atoms
        Calculated atomic coordinates w (float):
           Calculated w parameter of the atomic coordinates of all
           carbon atoms

    Sources:
        https://archive.ics.uci.edu/ml/datasets/Carbon+Nanotubes
        https://doi.org/10.1007/s00339-016-0153-1
        https://doi.org/10.17341/gazimmfd.337642

    Examples:
        Load in the data set::

            >>> dataset = Nanotube()
            >>> dataset.shape
            (10721, 8)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((10721, 5), (10721, 3))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((8541, 5), (8541, 3), (2180, 5), (2180, 3))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00448/carbon_nanotubes.csv"
    )

    _features = range(5)
    _targets = [5, 6, 7]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        csv_file = io.BytesIO(data)

        # Read the file-like object into a dataframe
        df = pd.read_csv(csv_file, sep=";", decimal=",")
        return df



"""Quantile regression for generalised linear models"""

from typing import Callable, Optional, Tuple, Union

import numpy as np
from scipy.optimize import minimize
from sklearn.linear_model import LinearRegression
from sklearn.linear_model._glm import _GeneralizedLinearRegressor
from sklearn.preprocessing import StandardScaler

from .quantile_loss import quantile_loss, smooth_quantile_loss


class QuantileRegressor:
    """Quantile regression for generalised linear models.

    This uses BFGS optimisation of the smooth quantile loss from [1].

    Args:
        max_iter (int):
            The maximal number of iterations to train the model for. Defaults to
            10,000.
        uncertainty (float):
            The uncertainty in the prediction intervals. Must be between 0 and 1.
            Defaults to 0.05.
        quantiles (sequence of floats or None, optional):
            List of quantiles to output, as an alternative to the `uncertainty`
            argument, and will not be used if that argument is set. If None then
            `uncertainty` is used. Defaults to None.
        alpha (float, optional):
            Smoothing parameter. Defaults to 0.4.

    Examples:
        Fitting and predicting follows scikit-learn syntax::

            >>> from doubt.datasets import Concrete
            >>> from sklearn.linear_model import PoissonRegressor
            >>> import numpy as np
            >>> X, y = Concrete().split(random_seed=42)
            >>> model = QuantileRegressor(PoissonRegressor(max_iter=10_000),
            ...                           uncertainty=0.05)
            >>> model.fit(X, y).predict(X)[0].shape
            (1030,)
            >>> x = [500, 0, 0, 100, 2, 1000, 500, 20]
            >>> preds, (start, end) = model.predict(x)
            >>> 78 < preds < 79
            True
            >>> 22 < start < 23
            True
            >>> 108 < end < 112
            True

    Sources:
        [1]: Songfeng Zheng (2011). Gradient Descent Algorithms for
             Quantile Regression With Smooth Approximation. International
             Journal of Machine Learning and Cybernetics.
    """

    def __init__(
        self,
        model: Union[LinearRegression, _GeneralizedLinearRegressor],
        max_iter: Optional[int] = None,
        uncertainty: float = 0.05,
        quantiles: Optional[np.ndarray] = None,
        alpha: float = 0.4,
    ):

        self.uncertainty = uncertainty
        self.alpha = alpha

        self._model = model
        self._scaler = StandardScaler()

        # Set `max_iter` to be the model's `max_iter` attribute if it exists, and
        # otherwise default to 10,000
        if max_iter is None:
            if hasattr(model, "max_iter"):
                self.max_iter = model.max_iter
            else:
                self.max_iter = 10_000
        else:
            self.max_iter = max_iter

        # Default quantiles to the prediction interval
        if quantiles is None:
            self.quantiles = [uncertainty / 2, 1 - (uncertainty / 2)]
        else:
            self.quantiles = list(quantiles)

        # Initialise inverse link function and weights
        self._inverse_link_function: Callable
        self._weights = {q: None for q in self.quantiles}

    def _objective_function(
        self,
        beta: np.ndarray,
        X: np.ndarray,
        y: np.ndarray,
        quantile: float,
        inverse_link_function: Callable,
    ) -> float:
        """Function used to optimise the quantile loss"""
        predictions = inverse_link_function(X @ beta)
        loss = smooth_quantile_loss(
            predictions=predictions, targets=y, quantile=quantile, alpha=self.alpha
        )
        return loss

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs):
        """Fit the model.

        Args:
            X (float matrix):
                The array containing the data set, either of shape (n,) or (n, f), with
                n being the number of samples and f being the number of features.
            y (float array):
                The target array, of shape (n,).
        """
        # Convert inputs to Numpy arrays
        X_arr = np.asarray(X)
        y_arr = np.asarray(y)

        # Fit the scaler and scale `X`
        self._scaler.fit(X_arr)
        X_arr = self._scaler.transform(X_arr)

        # Fit the underlying model
        self._model.fit(X_arr, y_arr)

        # Add constant feature
        X_arr = np.concatenate((X_arr, np.ones((X_arr.shape[0], 1))), axis=1)

        # When a Scikit-Learn GLM is fitted then its link function is set to the
        # attribute _link_instance. As the LinearRegression class does not subclass the
        # GeneralizedLinearRegressor class then we deal with this case separately
        try:
            link = self._model._link_instance
            self._inverse_link_function = link.inverse
        except AttributeError:
            self._inverse_link_function = lambda x: x

        # Use the underlying GLM's parameters for the initial guess of the quantile
        # weights
        model_weights = self._model.coef_
        model_intercept = self._model.intercept_
        beta_init = np.concatenate((model_weights, [model_intercept]))

        # Fit all quantile estimates
        for q in self.quantiles:

            args = (X_arr, y_arr, q, self._inverse_link_function)
            result = minimize(
                self._objective_function,
                beta_init,
                args=args,
                method="BFGS",
                options={"maxiter": self.max_iter},
            )
            self._weights[q] = result.x

        return self

    def predict(
        self, X: np.ndarray, **kwargs
    ) -> Tuple[Union[float, np.ndarray], np.ndarray]:
        """Compute model predictions.

        Args:
            X (float matrix):
                The array containing the data set, either of shape (n,) or (n, f), with
                n being the number of samples and f being the number of features.

        Returns:
            pair of float arrays:
                The predictions, of shape (n,), and the prediction intervals, of shape
                (n, 2).
        """
        # Convert inputs to Numpy array
        X_arr = np.asarray(X)

        # If input is one-dimensional, then add a dimension to it
        onedim = len(X_arr.shape) == 1
        if onedim:
            X_arr = np.expand_dims(X_arr, 0)

        # Standardise the feature matrix
        X_arr = self._scaler.transform(X_arr)

        # Compute the predictions from the underlying model
        preds = self._model.predict(X_arr)

        # Add constant feature
        X_arr = np.concatenate((X_arr, np.ones((X_arr.shape[0], 1))), axis=1)

        # Get the prediction for the lower- and upper quantiles
        quantile_vals = [
            self._inverse_link_function(X_arr @ self._weights[q])
            for q in self._weights.keys()
        ]

        # Concatenate the quantiles to get the intervals
        quantile_vals = np.stack(quantile_vals, axis=1).squeeze()

        # If we started with a one-dimensional input, ensure that the output is also
        # one-dimensional
        if onedim:
            preds = preds.item()

        return preds, quantile_vals

    def score(self, X: np.ndarray, y: np.ndarray) -> float:
        """Compute either the R^2 value or the negative pinball loss.

        If `uncertainty` is not set in the constructor then the R^2 value will be
        returned, and otherwise the mean of the two negative pinball losses
        corresponding to the two quantiles will be returned.

        The pinball loss is computed as quantile * (target - prediction) if target >=
        prediction, and (1 - quantile)(prediction - target) otherwise.

        Args:
            X (float array):
                The array containing the data set, either of shape (n,) or (n, f), with
                n being the number of samples and f being the number of features.
            y (float array):
                The target array, of shape (n,).

        Returns:
            float: The negative pinball loss.
        """
        # Convert inputs to Numpy arrays
        X_arr = np.asarray(X)
        y_arr = np.asarray(y)

        # If `uncertainty` has not been specified then simply use the `score` method
        # from the underlying LinearRegression model, which returns the R^2 value
        if self.uncertainty is None:
            return self._linreg.score(X_arr, y_arr)

        # If `uncertainty` has been specified then compute pinball loss
        else:
            # Get the predictions
            _, quantile_vals = self.predict(X_arr)
            losses = [
                quantile_loss(y_arr, quantile_vals[:, i], q)
                for i, q in enumerate(self.quantiles)
            ]
            return -np.mean(losses)

    def __repr__(self) -> str:
        model_name = self._model.__class__.__name__
        return (
            f"QuantileRegressor(model={model_name},\n"
            f"                  quantiles={self.quantiles})\n"
            f"                  alpha={self.alpha})\n"
            f"                  max_iter={self.max_iter}))"
        )



"""Fish toxicity data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class FishToxicity(BaseDataset):
    __doc__ = f"""
    This dataset was used to develop quantitative regression QSAR models to predict
    acute aquatic toxicity towards the fish Pimephales promelas (fathead minnow) on a
    set of 908 chemicals. LC50 data, which is the concentration that causes death in
    50% of test fish over a test duration of 96 hours, was used as model response

    {BASE_DATASET_DESCRIPTION}

    Features:
        CIC0 (float):
            Information indices
        SM1_Dz(Z) (float):
            2D matrix-based descriptors
        GATS1i (float):
            2D autocorrelations
        NdsCH (int)
            Atom-type counts
        NdssC (int)
            Atom-type counts
        MLOGP (float):
            Molecular properties

    Targets:
        LC50 (float):
            A concentration that causes death in 50% of test fish over a test duration
            of 96 hours. In -log(mol/L) units.

    Source:
        https://archive.ics.uci.edu/ml/datasets/QSAR+fish+toxicity

    Examples:
        Load in the data set::

            >>> dataset = FishToxicity()
            >>> dataset.shape
            (908, 7)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((908, 6), (908,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((708, 6), (708,), (200, 6), (200,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00504/qsar_fish_toxicity.csv"
    )

    _features = range(6)
    _targets = [6]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        csv_file = io.BytesIO(data)

        # Read the file-like object into a dataframe
        cols = ["CIC0", "SM1_Dz(Z)", "GATS1i", "NdsCH", "NdssC", "MLOGP", "LC50"]
        df = pd.read_csv(csv_file, sep=";", header=None, names=cols)

        return df



"""
.. include:: ../../README.md
"""

import pkg_resources

from .models import Boot  # noqa
from .models import QuantileRegressionForest  # noqa
from .models import QuantileRegressionTree  # noqa
from .models import QuantileRegressor  # noqa

# Fetches the version of the package as defined in pyproject.toml
__version__ = pkg_resources.get_distribution("doubt").version



"""Servo data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class Servo(BaseDataset):
    __doc__ = f"""
    Data was from a simulation of a servo system.

    Ross Quinlan:

    This data was given to me by Karl Ulrich at MIT in 1986. I didn't record his
    description at the time, but here's his subsequent (1992) recollection:

    "I seem to remember that the data was from a simulation of a servo system involving
    a servo amplifier, a motor, a lead screw/nut, and a sliding carriage of some sort.
    It may have been on of the translational axes of a robot on the 9th floor of the AI
    lab. In any case, the output value is almost certainly a rise time, or the time
    required for the system to respond to a step change in a position set point."

    (Quinlan, ML'93)

    "This is an interesting collection of data provided by Karl Ulrich. It covers an
    extremely non-linear phenomenon - predicting the rise time of a servomechanism in
    terms of two (continuous) gain settings and two (discrete) choices of mechanical
    linkages."

    {BASE_DATASET_DESCRIPTION}

    Features:
        motor (int):
            Motor, ranges from 0 to 4 inclusive
        screw (int):
            Screw, ranges from 0 to 4 inclusive
        pgain (int):
            PGain, ranges from 3 to 6 inclusive
        vgain (int):
            VGain, ranges from 1 to 5 inclusive

    Targets:
        class (float):
            Class values, ranges from 0.13 to 7.10 inclusive

    Source:
        https://archive.ics.uci.edu/ml/datasets/Servo

    Examples:
        Load in the data set::

            >>> dataset = Servo()
            >>> dataset.shape
            (167, 5)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((167, 4), (167,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((131, 4), (131,), (36, 4), (36,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/" "servo/servo.data"
    )

    _features = range(4)
    _targets = [4]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        csv_file = io.BytesIO(data)

        # Load in the dataframe
        cols = ["motor", "screw", "pgain", "vgain", "class"]
        df = pd.read_csv(csv_file, names=cols)

        # Encode motor and screw
        codes = ["A", "B", "C", "D", "E"]
        df["motor"] = df.motor.map(lambda x: codes.index(x))
        df["screw"] = df.screw.map(lambda x: codes.index(x))

        return df



"""Protocol class for estimators."""

from typing import Protocol, Tuple, Union

import numpy as np


class Model(Protocol):
    def __init__(self, *args, **kwargs):
        ...

    def predict(
        self, X: np.ndarray, **kwargs
    ) -> Tuple[Union[float, np.ndarray], np.ndarray]:
        ...

    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs):
        ...

    def __call__(
        self, X: np.ndarray, **kwargs
    ) -> Tuple[Union[float, np.ndarray], np.ndarray]:
        ...



"""Protein data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class Protein(BaseDataset):
    __doc__ = f"""
    This is a data set of Physicochemical Properties of Protein Tertiary Structure. The
    data set is taken from CASP 5-9. There are 45730 decoys and size varying from 0 to
    21 armstrong.

    {BASE_DATASET_DESCRIPTION}

    Features:
        F1 (float):
            Total surface area
        F2 (float):
            Non polar exposed area
        F3 (float):
            Fractional area of exposed non polar residue
        F4 (float):
            Fractional area of exposed non polar part of residue
        F5 (float):
            Molecular mass weighted exposed area
        F6 (float):
            Average deviation from standard exposed area of residue
        F7 (float):
            Euclidean distance
        F8 (float):
            Secondary structure penalty
        F9 (float):
            Spacial Distribution constraints (N,K Value)

    Targets:
        RMSD (float):
            Size of the residue

    Source:
        https://archive.ics.uci.edu/ml/datasets/Physicochemical+Properties+of+Protein+Tertiary+Structure

    Examples:
        Load in the data set::

            >>> dataset = Protein()
            >>> dataset.shape
            (45730, 10)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((45730, 9), (45730,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((36580, 9), (36580,), (9150, 9), (9150,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = "https://archive.ics.uci.edu/ml/machine-learning-databases/" "00265/CASP.csv"

    _features = range(9)
    _targets = [9]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        csv_file = io.BytesIO(data)

        # Load in the dataframe
        df = pd.read_csv(csv_file)

        # Put the target column at the end
        df = df[[f"F{i}" for i in range(1, 10)] + ["RMSD"]]

        return df



"""Facebook metrics data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io
import zipfile

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class FacebookMetrics(BaseDataset):
    __doc__ = f"""
    The data is related to posts' published during the year of 2014 on the Facebook's
    page of a renowned cosmetics brand.

    {BASE_DATASET_DESCRIPTION}

    Features:
        page_likes(int):
            The total number of likes of the Facebook page at the given time.
        post_type (int):
            The type of post. Here 0 means 'Photo', 1 means 'Status', 2 means 'Link'
            and 3 means 'Video'
        post_category (int):
            The category of the post.
        post_month (int):
            The month the post was posted, from 1 to 12 inclusive.
        post_weekday (int):
            The day of the week the post was posted, from 1 to 7 inclusive.
        post_hour (int):
            The hour the post was posted, from 0 to 23 inclusive
        paid (int):
            Binary feature, whether the post was paid for.

    Targets:
        total_reach (int):
            The lifetime post total reach.
        total_impressions (int):
            The lifetime post total impressions.
        engaged_users (int):
            The lifetime engaged users.
        post_consumers (int):
            The lifetime post consumers.
        post_consumptions (int):
            The lifetime post consumptions.
        post_impressions (int):
            The lifetime post impressions by people who liked the page.
        post_reach (int):
            The lifetime post reach by people who liked the page.
        post_engagements (int):
            The lifetime people who have liked the page and engaged with
            the post.
        comments (int):
            The number of comments.
        shares (int):
            The number of shares.
        total_interactions (int):
            The total number of interactions

    Source:
        https://archive.ics.uci.edu/ml/datasets/Facebook+metrics

    Examples:
        Load in the data set::

            >>> dataset = FacebookMetrics()
            >>> dataset.shape
            (500, 18)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((500, 7), (500, 11))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((388, 7), (388, 11), (112, 7), (112, 11))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00368/Facebook_metrics.zip"
    )

    _features = range(7)
    _targets = range(7, 18)

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        buffer = io.BytesIO(data)

        # Unzip the file and pull out the csv file
        with zipfile.ZipFile(buffer, "r") as zip_file:
            csv = zip_file.read("dataset_Facebook.csv")

        # Convert the bytes into a file-like object
        csv_file = io.BytesIO(csv)

        # Read the file-like object into a dataframe
        cols = [
            "page_likes",
            "post_type",
            "post_category",
            "post_month",
            "post_weekday",
            "post_hour",
            "paid",
            "total_reach",
            "total_impressions",
            "engaged_users",
            "post_consumers",
            "post_consumptions",
            "post_impressions",
            "post_reach",
            "post_engagements",
            "comments",
            "shares",
            "total_interactions",
        ]
        df = pd.read_csv(csv_file, sep=";", names=cols, header=0, index_col=0)

        # Numericalise post type
        post_types = list(df.post_type.unique())
        df["post_type"] = df.post_type.map(lambda txt: post_types.index(txt))

        return df



"""Fish bioconcentration data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io
import zipfile

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class FishBioconcentration(BaseDataset):
    __doc__ = f"""
    This dataset contains manually-curated experimental bioconcentration factor (BCF)
    for 1058 molecules (continuous values). Each row contains a molecule, identified by
    a CAS number, a name (if available), and a SMILES string. Additionally, the KOW
    (experimental or predicted) is reported. In this database, you will also find
    Extended Connectivity Fingerprints (binary vectors of 1024 bits), to be used as
    independent variables to predict the BCF.

    {BASE_DATASET_DESCRIPTION}

    Features:
        logkow (float):
            Octanol water paritioning coefficient (experimental or predicted, as
            indicated by ``KOW type``
        kow_exp (int):
            Indicates whether ``logKOW`` is experimental or predicted, with 1 denoting
            experimental and 0 denoting predicted
        smiles_[idx] for idx = 0..125 (int):
            Encoding of SMILES string to identify the 2D molecular structure. The
            encoding is as follows, where 'x' is a padding string to ensure that all
            the SMILES strings are of the same length:

                - 0  = 'x'
                - 1  = '#'
                - 2  = '('
                - 3  = ')'
                - 4  = '+'
                - 5  = '-'
                - 6  = '/'
                - 7  = '1'
                - 8  = '2'
                - 9  = '3'
                - 10 = '4'
                - 11 = '5'
                - 12 = '6'
                - 13 = '7'
                - 14 = '8'
                - 15 = '='
                - 16 = '@'
                - 17 = 'B'
                - 18 = 'C'
                - 19 = 'F'
                - 20 = 'H'
                - 21 = 'I'
                - 22 = 'N'
                - 23 = 'O'
                - 24 = 'P'
                - 25 = 'S'
                - 26 = '['
                - 27 = '\\'
                - 28 = ']'
                - 29 = 'c'
                - 30 = 'i'
                - 31 = 'l'
                - 32 = 'n'
                - 33 = 'o'
                - 34 = 'r'
                - 35 = 's'

    Targets:
        logbcf (float):
            Experimental fish bioconcentration factor (logarithm form)

    Source:
        https://archive.ics.uci.edu/ml/datasets/QSAR+fish+bioconcentration+factor+%28BCF%29

    Examples:
        Load in the data set::

            >>> dataset = FishBioconcentration()
            >>> dataset.shape
            (1054, 129)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((1054, 128), (1054,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((825, 128), (825,), (229, 128), (229,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00511/QSAR_fish_BCF.zip"
    )

    _features = range(128)
    _targets = [128]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        buffer = io.BytesIO(data)

        # Unzip the file and pull out the csv file
        with zipfile.ZipFile(buffer, "r") as zip_file:
            csv = zip_file.read("QSAR_BCF_Kow.csv")

        # Convert the string into a file-like object
        csv_file = io.BytesIO(csv)

        # Read the file-like object into a dataframe
        cols = ["cas", "name", "smiles", "logkow", "kow_exp", "logbcf"]
        df = pd.read_csv(
            csv_file,
            names=cols,
            header=0,
            usecols=[col for col in cols if col not in ["cas", "name"]],
        )

        # Drop NaNs
        df = df.dropna().reset_index(drop=True)

        # Encode KOW types
        kow_types = ["pred", "exp"]
        df["kow_exp"] = df.kow_exp.map(lambda txt: kow_types.index(txt))

        # Get maximum SMILE string length
        max_smile = max(len(smile_string) for smile_string in df.smiles)

        # Pad SMILE strings
        df["smiles"] = [
            smile_string + "x" * (max_smile - len(smile_string))
            for smile_string in df.smiles
        ]

        # Split up the SMILE strings into a matrix
        smile_df = pd.DataFrame(df.smiles.map(list).values.tolist())

        # Set the column values of the SMILE dataframe
        smile_df.columns = pd.Index(
            [f"smiles_{idx}" for idx in range(smile_df.shape[1])]
        )

        # Add the smile dataframe to the original dataframe
        df = pd.concat([df, smile_df], axis=1)

        # Drop original SMILE feature
        df = df.drop(columns="smiles")

        # Put the target variable at the end
        cols = ["logkow", "kow_exp"]
        cols += [f"smiles_{idx}" for idx in range(max_smile)]
        cols += ["logbcf"]
        df = df[cols]

        # Ensure that the `logkow` column is numeric
        df["logkow"] = pd.to_numeric(df.logkow)

        return df



"""Superconductivity data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io
import zipfile

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class Superconductivity(BaseDataset):
    __doc__ = f"""
    This dataset contains data on 21,263 superconductors and their relevant features.
    The goal here is to predict the critical temperature based on the features
    extracted.

    {BASE_DATASET_DESCRIPTION}

    Features:
        - number_of_elements (int)
        - mean_atomic_mass (float)
        - wtd_mean_atomic_mass (float)
        - gmean_atomic_mass (float)
        - wtd_gmean_atomic_mass (float)
        - entropy_atomic_mass (float)
        - wtd_entropy_atomic_mass (float)
        - range_atomic_mass (float)
        - wtd_range_atomic_mass (float)
        - std_atomic_mass (float)
        - wtd_std_atomic_mass (float)
        - mean_fie (float)
        - wtd_mean_fie (float)
        - gmean_fie (float)
        - wtd_gmean_fie (float)
        - entropy_fie (float)
        - wtd_entropy_fie (float)
        - range_fie (float)
        - wtd_range_fie (float)
        - std_fie (float)
        - wtd_std_fie (float)
        - mean_atomic_radius (float)
        - wtd_mean_atomic_radius (float)
        - gmean_atomic_radius (float)
        - wtd_gmean_atomic_radius (float)
        - entropy_atomic_radius (float)
        - wtd_entropy_atomic_radius (float)
        - range_atomic_radius (float)
        - wtd_range_atomic_radius (float)
        - std_atomic_radius (float)
        - wtd_std_atomic_radius (float)
        - mean_Density (float)
        - wtd_mean_Density (float)
        - gmean_Density (float)
        - wtd_gmean_Density (float)
        - entropy_Density (float)
        - wtd_entropy_Density (float)
        - range_Density (float)
        - wtd_range_Density (float)
        - std_Density (float)
        - wtd_std_Density (float)
        - mean_ElectronAffinity (float)
        - wtd_mean_ElectronAffinity (float)
        - gmean_ElectronAffinity (float)
        - wtd_gmean_ElectronAffinity (float)
        - entropy_ElectronAffinity (float)
        - wtd_entropy_ElectronAffinity (float)
        - range_ElectronAffinity (float)
        - wtd_range_ElectronAffinity (float)
        - std_ElectronAffinity (float)
        - wtd_std_ElectronAffinity (float)
        - mean_FusionHeat (float)
        - wtd_mean_FusionHeat (float)
        - gmean_FusionHeat (float)
        - wtd_gmean_FusionHeat (float)
        - entropy_FusionHeat (float)
        - wtd_entropy_FusionHeat (float)
        - range_FusionHeat (float)
        - wtd_range_FusionHeat (float)
        - std_FusionHeat (float)
        - wtd_std_FusionHeat (float)
        - mean_ThermalConductivity (float)
        - wtd_mean_ThermalConductivity (float)
        - gmean_ThermalConductivity (float)
        - wtd_gmean_ThermalConductivity (float)
        - entropy_ThermalConductivity (float)
        - wtd_entropy_ThermalConductivity (float)
        - range_ThermalConductivity (float)
        - wtd_range_ThermalConductivity (float)
        - std_ThermalConductivity (float)
        - wtd_std_ThermalConductivity (float)
        - mean_Valence (float)
        - wtd_mean_Valence (float)
        - gmean_Valence (float)
        - wtd_gmean_Valence (float)
        - entropy_Valence (float)
        - wtd_entropy_Valence (float)
        - range_Valence (float)
        - wtd_range_Valence (float)
        - std_Valence (float)
        - wtd_std_Valence (float)

    Targets:
        - critical_temp (float)

    Source:
        https://archive.ics.uci.edu/ml/datasets/Superconductivty+Data

    Examples:
        Load in the data set::

            >>> dataset = Superconductivity()
            >>> dataset.shape
            (21263, 82)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((21263, 81), (21263,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((17004, 81), (17004,), (4259, 81), (4259,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00464/superconduct.zip"
    )

    _features = range(81)
    _targets = [81]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        buffer = io.BytesIO(data)

        # Unzip the file and pull out the text
        with zipfile.ZipFile(buffer, "r") as zip_file:
            txt = zip_file.read("train.csv")

        # Convert text to csv file
        csv_file = io.BytesIO(txt)

        # Load the csv file into a dataframe
        df = pd.read_csv(csv_file)

        return df



from .airfoil import Airfoil  # noqa
from .bike_sharing_daily import BikeSharingDaily  # noqa
from .bike_sharing_hourly import BikeSharingHourly  # noqa
from .blog import Blog  # noqa
from .concrete import Concrete  # noqa
from .cpu import CPU  # noqa
from .facebook_comments import FacebookComments  # noqa
from .facebook_metrics import FacebookMetrics  # noqa
from .fish_bioconcentration import FishBioconcentration  # noqa
from .fish_toxicity import FishToxicity  # noqa
from .forest_fire import ForestFire  # noqa
from .gas_turbine import GasTurbine  # noqa
from .nanotube import Nanotube  # noqa
from .new_taipei_housing import NewTaipeiHousing  # noqa
from .parkinsons import Parkinsons  # noqa
from .power_plant import PowerPlant  # noqa
from .protein import Protein  # noqa
from .servo import Servo  # noqa
from .solar_flare import SolarFlare  # noqa
from .space_shuttle import SpaceShuttle  # noqa
from .stocks import Stocks  # noqa
from .superconductivity import Superconductivity  # noqa
from .tehran_housing import TehranHousing  # noqa
from .yacht import Yacht  # noqa



"""Parkinsons data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class Parkinsons(BaseDataset):
    __doc__ = f"""
    This dataset is composed of a range of biomedical voice measurements from 42 people
    with early-stage Parkinson's disease recruited to a six-month trial of a
    telemonitoring device for remote symptom progression monitoring. The recordings
    were automatically captured in the patient's homes.

    Columns in the table contain subject number, subject age, subject gender, time
    interval from baseline recruitment date, motor UPDRS, total UPDRS, and 16
    biomedical voice measures. Each row corresponds to one of 5,875 voice recording
    from these individuals. The main aim of the data is to predict the motor and total
    UPDRS scores ('motor_UPDRS' and 'total_UPDRS') from the 16 voice measures.

    {BASE_DATASET_DESCRIPTION}

    Features:
        subject# (int):
            Integer that uniquely identifies each subject
        age (int):
            Subject age
        sex (int):
            Binary feature. Subject sex, with 0 being male and 1 female
        test_time (float):
            Time since recruitment into the trial. The integer part is the
            number of days since recruitment
        Jitter(%) (float):
            Measure of variation in fundamental frequency
        Jitter(Abs) (float):
            Measure of variation in fundamental frequency
        Jitter:RAP (float):
            Measure of variation in fundamental frequency
        Jitter:PPQ5 (float):
            Measure of variation in fundamental frequency
        Jitter:DDP (float):
            Measure of variation in fundamental frequency
        Shimmer (float):
            Measure of variation in amplitude
        Shimmer(dB) (float):
            Measure of variation in amplitude
        Shimmer:APQ3 (float):
            Measure of variation in amplitude
        Shimmer:APQ5 (float):
            Measure of variation in amplitude
        Shimmer:APQ11 (float):
            Measure of variation in amplitude
        Shimmer:DDA (float):
            Measure of variation in amplitude
        NHR (float):
            Measure of ratio of noise to tonal components in the voice
        HNR (float):
            Measure of ratio of noise to tonal components in the voice
        RPDE (float):
            A nonlinear dynamical complexity measure
        DFA (float):
            Signal fractal scaling exponent
        PPE (float):
            A nonlinear measure of fundamental frequency variation

    Targets:
        motor_UPDRS (float):
            Clinician's motor UPDRS score, linearly interpolated
        total_UPDRS (float):
            Clinician's total UPDRS score, linearly interpolated

    Source:
        https://archive.ics.uci.edu/ml/datasets/Parkinsons+Telemonitoring

    Examples:
        Load in the data set::

            >>> dataset = Parkinsons()
            >>> dataset.shape
            (5875, 22)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((5875, 20), (5875, 2))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((4659, 20), (4659, 2), (1216, 20), (1216, 2))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "parkinsons/telemonitoring/parkinsons_updrs.data"
    )

    _features = range(20)
    _targets = [20, 21]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        csv_file = io.BytesIO(data)

        # Load in dataframe
        df = pd.read_csv(csv_file, header=0)

        # Put target columns at the end
        cols = [col for col in df.columns if col[-5:] != "UPDRS"]
        df = df[cols + ["motor_UPDRS", "total_UPDRS"]]

        return df



"""Space shuttle data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io
import re

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class SpaceShuttle(BaseDataset):
    __doc__ = f"""
    The motivation for collecting this database was the explosion of the USA Space
    Shuttle Challenger on 28 January, 1986. An investigation ensued into the
    reliability of the shuttle's propulsion system. The explosion was eventually traced
    to the failure of one of the three field joints on one of the two solid booster
    rockets. Each of these six field joints includes two O-rings, designated as primary
    and secondary, which fail when phenomena called erosion and blowby both occur.

    The night before the launch a decision had to be made regarding launch safety. The
    discussion among engineers and managers leading to this decision included concern
    that the probability of failure of the O-rings depended on the temperature t at
    launch, which was forecase to be 31 degrees F. There are strong engineering reasons
    based on the composition of O-rings to support the judgment that failure
    probability may rise monotonically as temperature drops. One other variable, the
    pressure s at which safety testing for field join leaks was performed, was
    available, but its relevance to the failure process was unclear.

    Draper's paper includes a menacing figure graphing the number of field joints
    experiencing stress vs. liftoff temperature for the 23 shuttle flights previous to
    the Challenger disaster. No previous liftoff temperature was under 53 degrees F.
    Although tremendous extrapolation must be done from the given data to assess risk
    at 31 degrees F, it is obvious even to the layman "to foresee the unacceptably high
    risk created by launching at 31 degrees F." For more information, see Draper (1993)
    or the other previous analyses.

    The task is to predict the number of O-rings that will experience thermal distress
    for a given flight when the launch temperature is below freezing.

    {BASE_DATASET_DESCRIPTION}

    Features:
        idx (int):
            Temporal order of flight
        temp (int):
            Launch temperature in Fahrenheit
        pres (int):
            Leak-check pressure in psi
        n_risky_rings (int):
            Number of O-rings at risk on a given flight

    Targets:
        n_distressed_rings (int):
            Number of O-rings experiencing thermal distress

    Source:
        https://archive.ics.uci.edu/ml/datasets/Challenger+USA+Space+Shuttle+O-Ring

    Examples:
        Load in the data set::

            >>> dataset = SpaceShuttle()
            >>> dataset.shape
            (23, 5)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((23, 4), (23,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((20, 4), (20,), (3, 4), (3,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "space-shuttle/o-ring-erosion-only.data"
    )

    _features = range(4)
    _targets = [4]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Collapse whitespace
        processed_data = re.sub(r" +", " ", data.decode("utf-8"))

        # Convert the bytes into a file-like object
        csv_file = io.StringIO(processed_data)

        # Load in dataframe
        cols = ["n_risky_rings", "n_distressed_rings", "temp", "pres", "idx"]
        df = pd.read_csv(csv_file, sep=" ", names=cols)

        # Reorder columns
        df = df[["idx", "temp", "pres", "n_risky_rings", "n_distressed_rings"]]

        return df



"""Quantile regression trees"""

from typing import Optional, Tuple, Union

import numpy as np
from sklearn.tree import BaseDecisionTree, DecisionTreeRegressor
from sklearn.utils import check_array, check_X_y

from .utils import weighted_percentile


class BaseTreeQuantileRegressor(BaseDecisionTree):
    def predict(
        self,
        X: np.ndarray,
        uncertainty: Optional[float] = None,
        quantiles: Optional[np.ndarray] = None,
        check_input: bool = True,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Predict regression value for X.

        Args:
            X (array-like or sparse matrix):
                The input samples, of shape [n_samples, n_features]. Internally, it
                will be converted to `dtype=np.float32` and if a sparse matrix is
                provided to a sparse `csr_matrix`.
            uncertainty (float or None, optional):
                Value ranging from 0 to 1. If None then no prediction intervals will be
                returned. Defaults to None.
            quantiles (sequence of floats or None, optional):
                List of quantiles to output, as an alternative to the `uncertainty`
                argument, and will not be used if that argument is set. If None then
                `uncertainty` is used. Defaults to None.
            check_input (boolean, optional):
                Allow to bypass several input checking. Don't use this parameter unless
                you know what you do. Defaults to True.

        Returns:
            Array or pair of arrays:
                Either array with predictions, of shape [n_samples,], or a pair of
                arrays with the first one being the predictions and the second one
                being the desired quantiles/intervals, of shape [n_samples, 2] if
                `uncertainty` is not None, and [n_samples, n_quantiles] if `quantiles`
                is not None.
        """
        # Apply method requires X to be of dtype np.float32
        X = check_array(X, dtype=np.float32, accept_sparse="csc")
        preds = super().predict(X, check_input=check_input)

        if uncertainty is not None or quantiles is not None:

            # Define list of quantiles that we want to output
            if uncertainty is not None:
                quantiles = [uncertainty / 2, 1 - (uncertainty / 2)]
            else:
                quantiles = list(quantiles or [])

            # Collect the leaves in the tree
            X_leaves = self.apply(X)
            unique_leaves = np.unique(X_leaves)

            # Initialise quantile values
            quantile_vals = np.empty((X.shape[0], len(quantiles)))

            # Populate the quantile values
            for leaf in unique_leaves:
                for idx, quantile in enumerate(quantiles):
                    X_leaf = X_leaves == leaf
                    y_leaf = self.y_train_[self.y_train_leaves_ == leaf]
                    quantile_vals[X_leaf, idx] = weighted_percentile(y_leaf, quantile)

            return preds, quantile_vals
        else:
            return preds

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
        check_input: bool = True,
    ):
        """Build a decision tree classifier from the training set (X, y).

        Args:
            X (array-like or sparse matrix)
                The training input samples, of shape [n_samples, n_features].
                Internally, it will be converted to `dtype=np.float32` and if a sparse
                matrix is provided to a sparse `csc_matrix`.
            y (array-like):
                The target values (class labels) as integers or strings, of shape
                [n_samples] or [n_samples, n_outputs].
            sample_weight (array-like or None, optional):
                Sample weights of shape = [n_samples]. If None, then samples are
                equally weighted. Splits that would create child nodes with net zero or
                negative weight are ignored while searching for a split in each node.
                Splits are also ignored if they would result in any single class
                carrying a negative weight in either child node. Defaults to None.
            check_input (boolean, optional):
                Allow to bypass several input checking. Don't use this parameter unless
                you know what you do. Defaults to True.
        """
        # y passed from a forest is 2-D. This is to silence the annoying
        # data-conversion warnings.
        y = np.asarray(y)
        if np.ndim(y) == 2 and y.shape[1] == 1:
            y = np.ravel(y)

        # Apply method requires X to be of dtype np.float32
        X, y = check_X_y(
            X, y, accept_sparse="csc", dtype=np.float32, multi_output=False
        )
        super().fit(
            X,
            y,
            sample_weight=sample_weight,
            check_input=check_input,
        )
        self.y_train_ = y

        # Stores the leaf nodes that the samples lie in.
        self.y_train_leaves_ = self.tree_.apply(X)
        return self


class QuantileRegressionTree(DecisionTreeRegressor, BaseTreeQuantileRegressor):
    """A decision tree regressor that provides quantile estimates.

    Args:
        criterion (string, optional):
            The function to measure the quality of a split. Supported criteria are
            'squared_error' for the mean squared error, which is equal to variance
            reduction as feature selection criterion, and 'absolute_error' for the mean
            absolute error. Defaults to 'squared_error'.
        splitter (string, optional):
            The strategy used to choose the split at each node. Supported strategies
            are 'best' to choose the best split and 'random' to choose the best random
            split. Defaults to 'best'.
        max_features (int, float, string or None, optional):
            The number of features to consider when looking for the best split:
            - If int, then consider `max_features` features at each split.
            - If float, then `max_features` is a percentage and `int(max_features *
              n_features)` features are considered at each split.
            - If 'auto', then `max_features=n_features`.
            - If 'sqrt', then `max_features=sqrt(n_features)`.
            - If 'log2', then `max_features=log2(n_features)`.
            - If None, then `max_features=n_features`.
            Note: the search for a split does not stop until at least one valid
            partition of the node samples is found, even if it requires to effectively
            inspect more than `max_features` features. Defaults to None.
        max_depth (int or None, optional):
            The maximum depth of the tree. If None, then nodes are expanded until all
            leaves are pure or until all leaves contain less than min_samples_split
            samples. Defaults to None.
        min_samples_split (int or float, optional):
            The minimum number of samples required to split an internal node:
            - If int, then consider `min_samples_split` as the minimum number.
            - If float, then `min_samples_split` is a percentage and
              `ceil(min_samples_split * n_samples)` are the minimum number of samples
              for each split. Defaults to 2.
        min_samples_leaf (int or float, optional):
            The minimum number of samples required to be at a leaf node:
            - If int, then consider `min_samples_leaf` as the minimum number.
            - If float, then `min_samples_leaf` is a percentage and
              `ceil(min_samples_leaf * n_samples)` are the minimum number of samples
              for each node. Defaults to 1.
        min_weight_fraction_leaf (float, optional):
            The minimum weighted fraction of the sum total of weights (of all the input
            samples) required to be at a leaf node. Samples have equal weight when
            sample_weight is not provided. Defaults to 0.0.
        max_leaf_nodes (int or None, optional):
            Grow a tree with `max_leaf_nodes` in best-first fashion. Best nodes are
            defined as relative reduction in impurity. If None then unlimited number of
            leaf nodes. Defaults to None.
        random_seed (int, RandomState instance or None, optional):
            If int, random_state is the seed used by the random number generator; If
            RandomState instance, random_state is the random number generator; If None,
            the random number generator is the RandomState instance used by
            `np.random`. Defaults to None.

    Attributes:
        feature_importances_ (array):
            The feature importances, of shape = [n_features]. The higher, the more
            important the feature. The importance of a feature is computed as the
            (normalized) total reduction of the criterion brought by that feature. It
            is also known as the Gini importance.
        max_features_ (int):
            The inferred value of max_features.
        n_features_ (int):
            The number of features when `fit` is performed.
        n_outputs_ (int):
            The number of outputs when `fit` is performed.
        tree_ (Tree object):
            The underlying Tree object.
        y_train_ (array-like):
            Train target values.
        y_train_leaves_ (array-like):
            Cache the leaf nodes that each training sample falls into.
            y_train_leaves_[i] is the leaf that y_train[i] ends up at.
    """

    def __init__(
        self,
        criterion: str = "squared_error",
        splitter: str = "best",
        max_features: Optional[Union[int, float, str]] = None,
        max_depth: Optional[int] = None,
        min_samples_split: Union[int, float] = 2,
        min_samples_leaf: Union[int, float] = 1,
        min_weight_fraction_leaf: float = 0.0,
        max_leaf_nodes: Optional[int] = None,
        random_seed: Union[int, np.random.RandomState, None] = None,
    ):
        super().__init__(
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            random_state=random_seed,
        )



"""Concrete data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class Concrete(BaseDataset):
    __doc__ = f"""
    Concrete is the most important material in civil engineering. The concrete
    compressive strength is a highly nonlinear function of age and ingredients.

    {BASE_DATASET_DESCRIPTION}

    Features:
        Cement (float):
            Kg of cement in an m3 mixture
        Blast Furnace Slag (float):
            Kg of blast furnace slag in an m3 mixture
        Fly Ash (float):
            Kg of fly ash in an m3 mixture
        Water (float):
            Kg of water in an m3 mixture
        Superplasticiser (float):
            Kg of superplasticiser in an m3 mixture
        Coarse Aggregate (float):
            Kg of coarse aggregate in an m3 mixture
        Fine Aggregate (float):
            Kg of fine aggregate in an m3 mixture
        Age (int):
            Age in days, between 1 and 365 inclusive

    Targets:
        Concrete Compressive Strength (float):
            Concrete compressive strength in megapascals

    Source:
        https://archive.ics.uci.edu/ml/datasets/Concrete+Compressive+Strength

    Examples:
        Load in the data set::

            >>> dataset = Concrete()
            >>> dataset.shape
            (1030, 9)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((1030, 8), (1030,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((807, 8), (807,), (223, 8), (223,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "concrete/compressive/Concrete_Data.xls"
    )

    _features = range(8)
    _targets = [8]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """

        # Convert the bytes into a file-like object
        xls_file = io.BytesIO(data)

        # Load the file-like object into a data frame
        df = pd.read_excel(xls_file)
        return df



"""Airfoil data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class Airfoil(BaseDataset):
    __doc__ = f"""
    The NASA data set comprises different size NACA 0012 airfoils at various wind
    tunnel speeds and angles of attack. The span of the airfoil and the observer
    position were the same in all of the experiments.

    {BASE_DATASET_DESCRIPTION}

    Features:
        int:
            Frequency, in Hertzs
        float:
            Angle of attack, in degrees
        float:
            Chord length, in meters
        float:
            Free-stream velocity, in meters per second
        float:
            Suction side displacement thickness, in meters

    Targets:
        float:
            Scaled sound pressure level, in decibels

    Source:
        https://archive.ics.uci.edu/ml/datasets/Airfoil+Self-Noise

    Examples:
        Load in the data set::

            >>> dataset = Airfoil()
            >>> dataset.shape
            (1503, 6)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((1503, 5), (1503,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((1181, 5), (1181,), (322, 5), (322,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00291/airfoil_self_noise.dat"
    )

    _features = range(5)
    _targets = [5]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        csv_file = io.BytesIO(data)

        # Read the file-like object into a data frame
        df = pd.read_csv(csv_file, sep="\t", header=None)
        return df



"""Solar flare data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class SolarFlare(BaseDataset):
    __doc__ = f"""
    Each class attribute counts the number of solar flares of a certain class that
    occur in a 24 hour period.

    The database contains 3 potential classes, one for the number of times a certain
    type of solar flare occured in a 24 hour period.

    Each instance represents captured features for 1 active region on the sun.

    The data are divided into two sections. The second section (flare.data2) has had
    much more error correction applied to the it, and has consequently been treated as
    more reliable.

    {BASE_DATASET_DESCRIPTION}

    Features:
        class (int):
            Code for class (modified Zurich class). Ranges from 0 to 6 inclusive
        spot_size (int):
            Code for largest spot size. Ranges from 0 to 5 inclusive
        spot_distr (int):
            Code for spot distribution. Ranges from 0 to 3 inclusive
        activity (int):
            Binary feature indicating 1 = reduced and 2 = unchanged
        evolution (int):
            0 = decay, 1 = no growth and 2 = growth
        flare_activity (int):
            Previous 24 hour flare activity code, where 0 = nothing as big as an M1, 1
            = one M1 and 2 = more activity than one M1
        is_complex (int):
            Binary feature indicating historically complex
        became_complex (int):
            Binary feature indicating whether the region became historically complex on
            this pass across the sun's disk
        large (int):
            Binary feature, indicating whether area is large
        large_spot (int):
            Binary feature, indicating whether the area of the largest spot is greater
            than 5

    Targets:
        C-class (int):
            C-class flares production by this region in the following 24 hours (common
            flares)
        M-class (int):
            M-class flares production by this region in the following 24 hours (common
            flares)
        X-class (int):
            X-class flares production by this region in the following 24 hours (common
            flares)

    Source:
        https://archive.ics.uci.edu/ml/datasets/Solar+Flare

    Examples:
        Load in the data set::

            >>> dataset = SolarFlare()
            >>> dataset.shape
            (1066, 13)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((1066, 10), (1066, 3))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((837, 10), (837, 3), (229, 10), (229, 3))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "solar-flare/flare.data2"
    )

    _features = range(10)
    _targets = range(10, 13)

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        csv_file = io.BytesIO(data)

        # Load in dataframe
        cols = [
            "class",
            "spot_size",
            "spot_distr",
            "activity",
            "evolution",
            "flare_activity",
            "is_complex",
            "became_complex",
            "large",
            "large_spot",
            "C-class",
            "M-class",
            "X-class",
        ]
        df = pd.read_csv(csv_file, sep=" ", skiprows=[0], names=cols)

        # Encode class
        encodings = ["A", "B", "C", "D", "E", "F", "H"]
        df["class"] = df["class"].map(lambda x: encodings.index(x))

        # Encode spot size
        encodings = ["X", "R", "S", "A", "H", "K"]
        df["spot_size"] = df.spot_size.map(lambda x: encodings.index(x))

        # Encode spot distribution
        encodings = ["X", "O", "I", "C"]
        df["spot_distr"] = df.spot_distr.map(lambda x: encodings.index(x))

        return df



"""Tehran housing data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class TehranHousing(BaseDataset):
    __doc__ = f"""
    Data set includes construction cost, sale prices, project variables, and economic
    variables corresponding to real estate single-family residential apartments in
    Tehran, Iran.

    {BASE_DATASET_DESCRIPTION}

    Features:
        start_year (int):
            Start year in the Persian calendar
        start_quarter (int)
            Start quarter in the Persian calendar
        completion_year (int)
            Completion year in the Persian calendar
        completion_quarter (int)
            Completion quarter in the Persian calendar
        V-1..V-8 (floats):
            Project physical and financial variables
        V-11-1..29-1 (floats):
            Economic variables and indices in time, lag 1
        V-11-2..29-2 (floats):
            Economic variables and indices in time, lag 2
        V-11-3..29-3 (floats):
            Economic variables and indices in time, lag 3
        V-11-4..29-4 (floats):
            Economic variables and indices in time, lag 4
        V-11-5..29-5 (floats):
            Economic variables and indices in time, lag 5

    Targets:
        construction_cost (float)
        sale_price (float)

    Source:
        https://archive.ics.uci.edu/ml/datasets/Residential+Building+Data+Set

    Examples:
        Load in the data set::

            >>> dataset = TehranHousing()
            >>> dataset.shape
            (371, 109)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((371, 107), (371, 2))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((288, 107), (288, 2), (83, 107), (83, 2))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00437/Residential-Building-Data-Set.xlsx"
    )

    _features = range(107)
    _targets = [107, 108]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        xlsx_file = io.BytesIO(data)

        # Load it into dataframe
        cols = (
            ["start_year", "start_quarter", "completion_year", "completion_quarter"]
            + [f"V-{i}" for i in range(1, 9)]
            + [f"V-{i}-{j}" for j in range(1, 6) for i in range(11, 30)]
            + ["construction_cost", "sale_price"]
        )
        df = pd.read_excel(xlsx_file, skiprows=[0, 1], names=cols)
        return df



"""Forest fire data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class ForestFire(BaseDataset):
    __doc__ = f"""
    This is a difficult regression task, where the aim is to predict the burned area of
    forest fires, in the northeast region of Portugal, by using meteorological and
    other data.

    {BASE_DATASET_DESCRIPTION}

    Features:
        X (float):
            The x-axis spatial coordinate within the Montesinho park map. Ranges from 1
            to 9.
        Y (float):
            The y-axis spatial coordinate within the Montesinho park map Ranges from 2
            to 9.
        month (int):
            Month of the year. Ranges from 0 to 11
        day (int):
            Day of the week. Ranges from 0 to 6
        FFMC (float):
            FFMC index from the FWI system. Ranges from 18.7 to 96.20
        DMC (float):
            DMC index from the FWI system. Ranges from 1.1 to 291.3
        DC (float):
            DC index from the FWI system. Ranges from 7.9 to 860.6
        ISI (float):
            ISI index from the FWI system. Ranges from 0.0 to 56.1
        temp (float):
            Temperature in Celsius degrees. Ranges from 2.2 to 33.3
        RH (float):
            Relative humidity in %. Ranges from 15.0 to 100.0
        wind (float):
            Wind speed in km/h. Ranges from 0.4 to 9.4
        rain (float):
            Outside rain in mm/m2. Ranges from 0.0 to 6.4

    Targets:
        area (float):
            The burned area of the forest (in ha). Ranges from 0.00 to 1090.84

    Notes:
        The target variable is very skewed towards 0.0, thus it may make sense to model
        with the logarithm transform.

    Source:
        https://archive.ics.uci.edu/ml/datasets/Forest+Fires

    Examples:
        Load in the data set::

            >>> dataset = ForestFire()
            >>> dataset.shape
            (517, 13)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((517, 12), (517,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((401, 12), (401,), (116, 12), (116,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "forest-fires/forestfires.csv"
    )

    _features = range(12)
    _targets = [12]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        csv_file = io.BytesIO(data)

        # Read the file-like object into a dataframe
        df = pd.read_csv(csv_file)

        # Encode month
        months = [
            "jan",
            "feb",
            "mar",
            "apr",
            "may",
            "jun",
            "jul",
            "aug",
            "sep",
            "oct",
            "nov",
            "dec",
        ]
        df["month"] = df.month.map(lambda string: months.index(string))

        # Encode day
        weekdays = ["mon", "tue", "wed", "thu", "fri", "sat", "sun"]
        df["day"] = df.day.map(lambda string: weekdays.index(string))

        return df



"""Blog post data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io
import zipfile

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class Blog(BaseDataset):
    __doc__ = f"""
    This data originates from blog posts. The raw HTML-documents of the blog posts were
    crawled and processed. The prediction task associated with the data is the
    prediction of the number of comments in the upcoming 24 hours. In order to simulate
    this situation, we choose a basetime (in the past) and select the blog posts that
    were published at most 72 hours before the selected base date/time. Then, we
    calculate all the features of the selected blog posts from the information that was
    available at the basetime, therefore each instance corresponds to a blog post. The
    target is the number of comments that the blog post received in the next 24 hours
    relative to the basetime.

    In the train data, the basetimes were in the years 2010 and 2011. In the test data
    the basetimes were in February and March 2012. This simulates the real-world
    situtation in which training data from the past is available to predict events in
    the future.

    The train data was generated from different basetimes that may temporally overlap.
    Therefore, if you simply split the train into disjoint partitions, the underlying
    time intervals may overlap. Therefore, the you should use the provided, temporally
    disjoint train and test splits in order to ensure that the evaluation is fair.

    {BASE_DATASET_DESCRIPTION}

    Features:
        Features 0-49 (float):
            50 features containing the average, standard deviation, minimum, maximum
            and median of feature 50-59 for the source of the current blog post, by
            which we mean the blog on which the post appeared. For example,
            myblog.blog.org would be the source of the post
            myblog.blog.org/post_2010_09_10
        Feature 50 (int):
            Total number of comments before basetime
        Feature 51 (int):
            Number of comments in the last 24 hours before the basetime
        Feature 52 (int):
            If T1 is the datetime 48 hours before basetime and T2 is the datetime 24
            hours before basetime, then this is the number of comments in the time
            period between T1 and T2
        Feature 53 (int):
            Number of comments in the first 24 hours after the publication of the blog
            post, but before basetime
        Feature 54 (int):
            The difference between Feature 51 and Feature 52
        Features 55-59 (int):
            The same thing as Features 50-51, but for links (trackbacks) instead of
            comments
        Feature 60 (float):
            The length of time between the publication of the blog post and basetime
        Feature 61 (int):
            The length of the blog post
        Features 62-261 (int):
            The 200 bag of words features for 200 frequent words of the text of the
            blog post
        Features 262-268 (int):
            Binary indicators for the weekday (Monday-Sunday) of the basetime
        Features 269-275 (int):
            Binary indicators for the weekday (Monday-Sunday) of the date of
            publication of the blog post
        Feature 276 (int):
            Number of parent pages: we consider a blog post P as a parent of blog post
            B if B is a reply (trackback) to P
        Features 277-279 (float):
            Minimum, maximum and average of the number of comments the parents received

    Targets:
        int:
            The number of comments in the next 24 hours (relative to baseline)

    Source:
        https://archive.ics.uci.edu/ml/datasets/BlogFeedback

    Examples:
        Load in the data set::

            >>> dataset = Blog()
            >>> dataset.shape
            (52397, 281)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((52397, 279), (52397,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((41949, 279), (41949,), (10448, 279), (10448,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00304/BlogFeedback.zip"
    )

    _features = range(279)
    _targets = [279]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        buffer = io.BytesIO(data)

        # Unzip the file and pull out blogData_train.csv as a string
        with zipfile.ZipFile(buffer, "r") as zip_file:
            csv = zip_file.read("blogData_train.csv").decode("utf-8")

        # Convert the string into a file-like object
        csv_file = io.StringIO(csv)

        # Read the file-like object into a dataframe
        df = pd.read_csv(csv_file, header=None)
        return df



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['doubt',
 'doubt.datasets',
 'doubt.models',
 'doubt.models.boot',
 'doubt.models.glm',
 'doubt.models.tree']

package_data = \
{'': ['*']}

install_requires = \
['joblib>=1.1.0,<2.0.0',
 'numpy>=1.23.0,<2.0.0',
 'openpyxl>=3.0.10,<4.0.0',
 'pandas>=1.4.0,<2.0.0',
 'scikit-learn>=1.1.1,<2.0.0',
 'tables>=3.7.0,<4.0.0',
 'tqdm>=4.62.0,<5.0.0',
 'xlrd>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'doubt',
    'version': '4.3.0',
    'description': 'Bringing back uncertainty to machine learning.',
    'long_description': '# Doubt\n\n*Bringing back uncertainty to machine learning.*\n\n______________________________________________________________________\n[![PyPI Status](https://badge.fury.io/py/doubt.svg)](https://pypi.org/project/doubt/)\n[![Documentation](https://img.shields.io/badge/docs-passing-green)](https://saattrupdan.github.io/doubt/doubt.html)\n[![License](https://img.shields.io/github/license/saattrupdan/doubt)](https://github.com/saattrupdan/doubt/blob/main/LICENSE)\n[![LastCommit](https://img.shields.io/github/last-commit/saattrupdan/doubt)](https://github.com/saattrupdan/doubt/commits/main)\n[![Code Coverage](https://img.shields.io/badge/Coverage-67%25-yellow.svg)](https://github.com/saattrupdan/doubt/tree/dev/tests)\n\nA Python package to include prediction intervals in the predictions of machine\nlearning models, to quantify their uncertainty.\n\n\n## Installation\n\nIf you do not already have HDF5 installed, then start by installing that. On MacOS this\ncan be done using `sudo port install hdf5` after\n[MacPorts](https://www.macports.org/install.php) have been installed. On Ubuntu you can\nget HDF5 with `sudo apt-get install python-dev python3-dev libhdf5-serial-dev`. After\nthat, you can install `doubt` with `pip`:\n\n```shell\npip install doubt\n```\n\n\n## Features\n\n- Bootstrap wrapper for all Scikit-Learn models\n    - Can also be used to calculate usual bootstrapped statistics of a dataset\n- Quantile Regression for all generalised linear models\n- Quantile Regression Forests\n- A uniform dataset API, with 24 regression datasets and counting\n\n\n## Quick Start\n\nIf you already have a model in Scikit-Learn, then you can simply\nwrap it in a `Boot` to enable predicting with prediction intervals:\n\n```python\n>>> from sklearn.linear_model import LinearRegression\n>>> from doubt import Boot\n>>> from doubt.datasets import PowerPlant\n>>>\n>>> X, y = PowerPlant().split()\n>>> clf = Boot(LinearRegression())\n>>> clf = clf.fit(X, y)\n>>> clf.predict([10, 30, 1000, 50], uncertainty=0.05)\n(481.9203102126274, array([473.43314309, 490.0313962 ]))\n```\n\nAlternatively, you can use one of the standalone models with uncertainty\noutputs. For instance, a `QuantileRegressionForest`:\n\n```python\n>>> from doubt import QuantileRegressionForest as QRF\n>>> from doubt.datasets import Concrete\n>>> import numpy as np\n>>>\n>>> X, y = Concrete().split()\n>>> clf = QRF(max_leaf_nodes=8)\n>>> clf.fit(X, y)\n>>> clf.predict(np.ones(8), uncertainty=0.25)\n(16.933590347847982, array([ 8.93456428, 26.0664534 ]))\n```\n',
    'author': 'Dan Saattrup Nielsen',
    'author_email': 'saattrupdan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/saattrupdan/doubt',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""Gas turbine data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io
import zipfile

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class GasTurbine(BaseDataset):
    __doc__ = f"""
    Data have been generated from a sophisticated simulator of a Gas Turbines (GT),
    mounted on a Frigate characterized by a COmbined Diesel eLectric And Gas (CODLAG)
    propulsion plant type.

    The experiments have been carried out by means of a numerical simulator of a naval
    vessel (Frigate) characterized by a Gas Turbine (GT) propulsion plant. The
    different blocks forming the complete simulator (Propeller, Hull, GT, Gear Box and
    Controller) have been developed and fine tuned over the year on several similar
    real propulsion plants. In view of these observations the available data are in
    agreement with a possible real vessel.

    In this release of the simulator it is also possible to take into account the
    performance decay over time of the GT components such as GT compressor and
    turbines.

    The propulsion system behaviour has been described with this parameters:

        - Ship speed (linear function of the lever position lp).
        - Compressor degradation coefficient kMc.
        - Turbine degradation coefficient kMt.

    so that each possible degradation state can be described by a combination of this
    triple (lp,kMt,kMc).

    The range of decay of compressor and turbine has been sampled with an uniform grid
    of precision 0.001 so to have a good granularity of representation.

    In particular for the compressor decay state discretization the kMc coefficient has
    been investigated in the domain [1; 0.95], and the turbine coefficient in the
    domain [1; 0.975].

    Ship speed has been investigated sampling the range of feasible speed from 3 knots
    to 27 knots with a granularity of representation equal to tree knots.

    A series of measures (16 features) which indirectly represents of the state of the
    system subject to performance decay has been acquired and stored in the dataset
    over the parameter's space.

    {BASE_DATASET_DESCRIPTION}

    Features:
        lever_position (float)
            The position of the lever
        ship_speed (float):
            The ship speed, in knots
        shaft_torque (float):
            The shaft torque of the gas turbine, in kN m
        turbine_revolution_rate (float):
            The gas turbine rate of revolutions, in rpm
        generator_revolution_rate (float):
            The gas generator rate of revolutions, in rpm
        starboard_propeller_torque (float):
            The torque of the starboard propeller, in kN
        port_propeller_torque (float):
            The torque of the port propeller, in kN
        turbine_exit_temp (float):
            Height pressure turbine exit temperature, in celcius
        inlet_temp (float):
            Gas turbine compressor inlet air temperature, in celcius
        outlet_temp (float):
            Gas turbine compressor outlet air temperature, in celcius
        turbine_exit_pres (float):
            Height pressure turbine exit pressure, in bar
        inlet_pres (float):
            Gas turbine compressor inlet air pressure, in bar
        outlet_pres (float):
            Gas turbine compressor outlet air pressure, in bar
        exhaust_pres (float):
            Gas turbine exhaust gas pressure, in bar
        turbine_injection_control (float):
            Turbine injection control, in percent
        fuel_flow (float):
            Fuel flow, in kg/s

    Targets:
        compressor_decay (type):
            Gas turbine compressor decay state coefficient
        turbine_decay (type):
            Gas turbine decay state coefficient

    Source:
        https://archive.ics.uci.edu/ml/datasets/Condition+Based+Maintenance+of+Naval+Propulsion+Plants

    Examples:
        Load in the data set::

            >>> dataset = GasTurbine()
            >>> dataset.shape
            (11934, 18)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((11934, 16), (11934, 2))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((9516, 16), (9516, 2), (2418, 16), (2418, 2))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00316/UCI%20CBM%20Dataset.zip"
    )

    _features = range(16)
    _targets = [16, 17]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        buffer = io.BytesIO(data)

        # Unzip the file and pull out the txt file
        with zipfile.ZipFile(buffer, "r") as zip_file:
            txt_bytes = zip_file.read("UCI CBM Dataset/data.txt")

        # Decode text and replace initial space on each line
        txt = txt_bytes[3:].decode("utf-8").replace("\n   ", "\n")

        # Convert the remaining triple spaces into commas, to make loading it as a csv
        # file easier
        txt = txt.replace("   ", ",")

        # Convert the string into a file-like object
        csv_file = io.StringIO(txt)

        # Read the file-like object into a dataframe
        cols = [
            "lever_position",
            "ship_speed",
            "shaft_torque",
            "turbine_revolution_rate",
            "generator_revolution_rate",
            "starboard_propeller_torque",
            "port_propeller_torque",
            "turbine_exit_temp",
            "inlet_temp",
            "outlet_temp",
            "turbine_exit_pres",
            "inlet_pres",
            "outlet_pres",
            "exhaust_pres",
            "turbine_injection_control",
            "fuel_flow",
            "compressor_decay",
            "turbine_decay",
        ]
        df = pd.read_csv(csv_file, header=None, names=cols)

        return df



"""Hourly bike sharing data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io
import zipfile

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class BikeSharingHourly(BaseDataset):
    __doc__ = f"""
    Bike sharing systems are new generation of traditional bike rentals where whole
    process from membership, rental and return back has become automatic. Through these
    systems, user is able to easily rent a bike from a particular position and return
    back at another position. Currently, there are about over 500 bike-sharing programs
    around the world which is composed of over 500 thousands bicycles. Today, there
    exists great interest in these systems due to their important role in traffic,
    environmental and health issues.

    Apart from interesting real world applications of bike sharing systems, the
    characteristics of data being generated by these systems make them attractive for
    the research. Opposed to other transport services such as bus or subway, the
    duration of travel, departure and arrival position is explicitly recorded in these
    systems. This feature turns bike sharing system into a virtual sensor network that
    can be used for sensing mobility in the city. Hence, it is expected that most of
    important events in the city could be detected via monitoring these data.

    {BASE_DATASET_DESCRIPTION}

    Features:
        instant (int):
            Record index
        season (int):
            The season, with 1 = winter, 2 = spring, 3 = summer and 4 = autumn
        yr (int):
            The year, with 0 = 2011 and 1 = 2012
        mnth (int):
            The month, from 1 to 12 inclusive
        hr (int):
            The hour of the day, from 0 to 23 inclusive
        holiday (int):
            Whether day is a holiday or not, binary valued
        weekday (int):
            The day of the week, from 0 to 6 inclusive
        workingday (int):
            Working day, 1 if day is neither weekend nor holiday, otherwise 0
        weathersit (int):
            Weather, encoded as

            1. Clear, few clouds, partly cloudy
            2. Mist and cloudy, mist and broken clouds, mist and few clouds
            3. Light snow, light rain and thunderstorm and scattered clouds, light rain
            and scattered clouds
            4. Heavy rain and ice pallets and thunderstorm and mist, or snow and fog

        temp (float):
            Max-min normalised temperature in Celsius, from -8 to +39
        atemp (float):
            Max-min normalised feeling temperature in Celsius, from -16 to +50
        hum (float):
            Scaled max-min normalised humidity, from 0 to 1
        windspeed (float):
            Scaled max-min normalised wind speed, from 0 to 1

    Targets:
        casual (int):
            Count of casual users
        registered (int):
            Count of registered users
        cnt (int):
            Sum of casual and registered users

    Source:
        https://archive.ics.uci.edu/ml/datasets/Bike+Sharing+Dataset

    Examples:
        Load in the data set::

            >>> dataset = BikeSharingHourly()
            >>> dataset.shape
            (17379, 16)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((17379, 13), (17379, 3))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((13873, 13), (13873, 3), (3506, 13), (3506, 3))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00275/Bike-Sharing-Dataset.zip"
    )

    _features = range(13)
    _targets = [13, 14, 15]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        buffer = io.BytesIO(data)

        # Unzip the file and pull out hour.csv as a string
        with zipfile.ZipFile(buffer, "r") as zip_file:
            csv = zip_file.read("hour.csv").decode("utf-8")

        # Convert the string into a file-like object
        csv_file = io.StringIO(csv)

        # Read the file-like object into a dataframe
        cols = [0] + list(range(2, 17))
        df = pd.read_csv(csv_file, usecols=cols)
        return df



"""Yacht data set.

This data set is from the UCI data set archive, with the description being the original
description verbatim. Some feature names may have been altered, based on the
description.
"""

import io

import pandas as pd

from .dataset import BASE_DATASET_DESCRIPTION, BaseDataset


class Yacht(BaseDataset):
    __doc__ = f"""
    Prediction of residuary resistance of sailing yachts at the initial design stage is
    of a great value for evaluating the ship's performance and for estimating the
    required propulsive power. Essential inputs include the basic hull dimensions and
    the boat velocity.

    The Delft data set comprises 251 full-scale experiments, which were performed at
    the Delft Ship Hydromechanics Laboratory for that purpose.

    These experiments include 22 different hull forms, derived from a parent form
    closely related to the "Standfast 43" designed by Frans Maas.

    {BASE_DATASET_DESCRIPTION}

    Features:
        pos (float):
            Longitudinal position of the center of buoyancy, adimensional
        prismatic (float):
            Prismatic coefficient, adimensional
        displacement (float):
            Length-displacement ratio, adimensional
        beam_draught (float):
            Beam-draught ratio, adimensional
        length_beam (float):
            Length-beam ratio, adimensional
        froude_no (float):
            Froude number, adimensional

    Targets:
        resistance (float):
            Residuary resistance per unit weight of displacement, adimensional

    Source:
        https://archive.ics.uci.edu/ml/datasets/Yacht+Hydrodynamics

    Examples:
        Load in the data set::

            >>> dataset = Yacht()
            >>> dataset.shape
            (251, 7)

        Split the data set into features and targets, as NumPy arrays::

            >>> X, y = dataset.split()
            >>> X.shape, y.shape
            ((251, 6), (251,))

        Perform a train/test split, also outputting NumPy arrays::

            >>> train_test_split = dataset.split(test_size=0.2, random_seed=42)
            >>> X_train, X_test, y_train, y_test = train_test_split
            >>> X_train.shape, y_train.shape, X_test.shape, y_test.shape
            ((196, 6), (196,), (55, 6), (55,))

        Output the underlying Pandas DataFrame::

            >>> df = dataset.to_pandas()
            >>> type(df)
            <class 'pandas.core.frame.DataFrame'>
    """

    _url = (
        "https://archive.ics.uci.edu/ml/machine-learning-databases/"
        "00243/yacht_hydrodynamics.data"
    )

    _features = range(6)
    _targets = [6]

    def _prep_data(self, data: bytes) -> pd.DataFrame:
        """Prepare the data set.

        Args:
            data (bytes): The raw data

        Returns:
            Pandas dataframe: The prepared data
        """
        # Convert the bytes into a file-like object
        txt_file = io.BytesIO(data)

        # Load it into dataframe
        cols = [
            "pos",
            "prismatic",
            "displacement",
            "beam_draught",
            "length_beam",
            "froude_no",
            "resistance",
        ]
        df = pd.read_csv(
            txt_file,
            header=None,
            sep=" ",
            names=cols,
            on_bad_lines="skip",
        )
        return df
